from __future__ import annotations

import subprocess
from types import SimpleNamespace
from typing import Any

import pytest

from mycli import vault


def test_get_field_from_vault_runs_kv_get_with_field_mount_and_address(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    run_calls: list[dict[str, Any]] = []

    def fake_run(command: list[str], **kwargs: Any) -> SimpleNamespace:
        run_calls.append({'command': command, **kwargs})
        return SimpleNamespace(returncode=0, stdout='secret\n', stderr='')

    monkeypatch.setattr(vault.subprocess, 'run', fake_run)

    password = vault.get_field_from_vault(
        'mysql_password',
        'database/prod',
        executable='/opt/bin/vault',
        mount='kv',
        address='https://vault.example.com',
    )

    assert password == 'secret'
    assert run_calls == [
        {
            'command': [
                '/opt/bin/vault',
                'kv',
                'get',
                '-field=mysql_password',
                '-mount=kv',
                '-address=https://vault.example.com',
                'database/prod',
            ],
            'check': False,
            'stdin': subprocess.DEVNULL,
            'stdout': subprocess.PIPE,
            'stderr': subprocess.PIPE,
            'text': True,
        }
    ]


def test_get_field_from_vault_reports_missing_executable(monkeypatch: pytest.MonkeyPatch) -> None:
    def fake_run(*_args: Any, **_kwargs: Any) -> SimpleNamespace:
        raise FileNotFoundError()

    monkeypatch.setattr(vault.subprocess, 'run', fake_run)

    with pytest.raises(vault.VaultError, match='Vault executable not found: missing-vault'):
        vault.get_field_from_vault('password', 'database/prod', executable='missing-vault')


def test_get_field_from_vault_reports_oserror(monkeypatch: pytest.MonkeyPatch) -> None:
    def fake_run(*_args: Any, **_kwargs: Any) -> SimpleNamespace:
        raise OSError('boom')

    monkeypatch.setattr(vault.subprocess, 'run', fake_run)

    with pytest.raises(vault.VaultError, match='Unable to run Vault executable vault: boom'):
        vault.get_field_from_vault('password', 'database/prod')


def test_get_field_from_vault_reports_nonzero_exit_without_stdout(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    def fake_run(*_args: Any, **_kwargs: Any) -> SimpleNamespace:
        return SimpleNamespace(returncode=2, stdout='secret\n', stderr='permission denied\n')

    monkeypatch.setattr(vault.subprocess, 'run', fake_run)

    with pytest.raises(vault.VaultError) as excinfo:
        vault.get_field_from_vault('password', 'database/prod')

    assert 'permission denied' in str(excinfo.value)
    assert 'secret' not in str(excinfo.value)


def test_get_field_from_vault_reports_nonzero_exit_without_stderr(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    def fake_run(*_args: Any, **_kwargs: Any) -> SimpleNamespace:
        return SimpleNamespace(returncode=2, stdout='', stderr='')

    monkeypatch.setattr(vault.subprocess, 'run', fake_run)

    with pytest.raises(vault.VaultError, match='Vault command failed\\. You may need to run "vault login"\\. Exit code 2\\.'):
        vault.get_field_from_vault('password', 'database/prod')



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any

from configobj import ConfigObj

from mycli.config import strip_matching_quotes
from mycli.types import ImageProtocol

if TYPE_CHECKING:
    from mycli.client import MyCli


def normalize_ssl_mode(
    config: ConfigObj,
    config_without_package_defaults: ConfigObj,
) -> tuple[str | None, str | None]:
    error_notice: str | None = None
    ssl_mode: str | None = None

    if 'main' in config_without_package_defaults and 'ssl_mode' in config_without_package_defaults['main']:
        # migration with notice added with mycli 2.0.0 in 2026-07
        # todo: entirely remove support for ssl_mode in [main]
        error_notice = (
            'Mycli 2.0 migration: automatically moving ssl_mode under [main] to default_ssl_mode under [connection] in ~/.myclirc .'
        )

        ssl_mode = config_without_package_defaults['main']['ssl_mode']

        config_without_package_defaults.encoding = 'utf-8'
        if 'connection' not in config_without_package_defaults:
            config_without_package_defaults['connection'] = {}
        if config_without_package_defaults['connection'].get('default_ssl_mode', None) in (None, ''):
            config_without_package_defaults['connection']['default_ssl_mode'] = ssl_mode
        else:
            ssl_mode = config_without_package_defaults['connection'].get('default_ssl_mode')
            error_notice += f'\nBut connection.default_ssl_mode already existed, with the value: "{ssl_mode}".'
        del config_without_package_defaults['main']['ssl_mode']
        config_without_package_defaults.write()

    if not ssl_mode and 'default_ssl_mode' in config['connection']:
        ssl_mode = config['connection']['default_ssl_mode']
    if ssl_mode not in ('auto', 'on', 'off', None):
        error_notice = f'Invalid config option provided for ssl_mode ({ssl_mode}); ignoring.'
        return None, error_notice
    return ssl_mode, error_notice


def normalize_image_protocol(image_protocol: str | None) -> tuple[ImageProtocol, str | None]:
    if image_protocol == 'iterm2':
        return 'iterm2', None
    if image_protocol == 'kitty':
        return 'kitty', None
    if image_protocol in ('none', '', None):
        return 'none', None
    return 'none', f'Invalid config option provided for image_protocol ({image_protocol}); disabling.'


def configure_prompt_state(
    mycli: MyCli,
    config: ConfigObj,
    prompt: str | None,
    toolbar_format: str | None,
) -> None:
    mycli.prompt_format = prompt or config['main']['prompt'] or mycli.default_prompt
    mycli.prompt_lines = 0
    mycli.multiline_continuation_char = config['main']['prompt_continuation']
    mycli.toolbar_format = toolbar_format or config['main']['toolbar']
    mycli.terminal_tab_title_format = config['main']['terminal_tab_title']
    mycli.terminal_window_title_format = config['main']['terminal_window_title']
    mycli.multiplex_window_title_format = config['main']['multiplex_window_title']
    mycli.multiplex_pane_title_format = config['main']['multiplex_pane_title']


def destructive_keywords_from_config(config: ConfigObj) -> list[str]:
    keywords = config['main'].get('destructive_keywords', 'DROP SHUTDOWN DELETE TRUNCATE ALTER UPDATE')
    return [keyword for keyword in keywords.split(' ') if keyword]


def llm_prompt_truncation(config: ConfigObj) -> tuple[int, int]:
    if 'llm' in config and re.match(r'^\d+$', config['llm'].get('prompt_field_truncate', '')):
        field_truncate = int(config['llm'].get('prompt_field_truncate'))
    else:
        field_truncate = 0
    if 'llm' in config and re.match(r'^\d+$', config['llm'].get('prompt_section_truncate', '')):
        section_truncate = int(config['llm'].get('prompt_section_truncate'))
    else:
        section_truncate = 0
    return field_truncate, section_truncate


class AppStateMixin:
    login_path: str | None

    def read_mylogin_cnf(self, cnf: ConfigObj) -> dict[str, Any]:
        allowed_keys = [
            'user',
            'password',
            'host',
            'port',
            'socket',
        ]
        configuration: dict[str, Any] = dict.fromkeys(allowed_keys)
        for section in cnf:
            if section != self.login_path:
                continue
            for key in allowed_keys:
                if key in cnf[section]:
                    configuration[key] = strip_matching_quotes(cnf[section][key])

        return configuration



from __future__ import annotations

import subprocess

DEFAULT_VAULT_EXECUTABLE = 'vault'
DEFAULT_VAULT_PASSWORD_FIELD = 'password'
DEFAULT_VAULT_USERNAME_FIELD = 'username'


class VaultError(RuntimeError):
    pass


def get_field_from_vault(
    field: str,
    secret: str,
    executable: str = DEFAULT_VAULT_EXECUTABLE,
    mount: str | None = None,
    address: str | None = None,
) -> str:
    command = [
        executable,
        'kv',
        'get',
        f'-field={field}',
    ]
    if mount:
        command.append(f'-mount={mount}')
    if address:
        command.append(f'-address={address}')

    command.append(secret)

    try:
        completed_process = subprocess.run(
            command,
            check=False,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
    except FileNotFoundError as exc:
        raise VaultError(f'Vault executable not found: {executable}') from exc
    except OSError as exc:
        raise VaultError(f'Unable to run Vault executable {executable}: {exc}') from exc

    if completed_process.returncode:
        stderr = completed_process.stderr.strip()
        if stderr:
            raise VaultError(f'Vault command failed. You may need to run "vault login": {stderr}')
        raise VaultError(f'Vault command failed. You may need to run "vault login". Exit code {completed_process.returncode}.')

    return completed_process.stdout.removesuffix('\n')



from types import SimpleNamespace
from typing import Any, cast

from mycli import client_query, main
from mycli.packages.sqlresult import SQLResult
from mycli.types import Query
from test.utils import (  # type: ignore[attr-defined]
    FakeCursorBase,
    ReusableLock,
    make_bare_mycli,
)


def make_refresh_cli() -> tuple[Any, dict[str, Any]]:
    cli = make_bare_mycli()
    state: dict[str, Any] = {
        'stopped': [],
        'refresh_calls': [],
        'set_dbname_calls': [],
    }
    callback = object()
    cli.schema_prefetcher = SimpleNamespace(stop=lambda: state['stopped'].append(True))
    cli.sqlexecute = SimpleNamespace(dbname='current')
    cli._on_completions_refreshed = callback
    cli.completer = SimpleNamespace(
        keyword_casing='upper',
        indexed_column_suffix=' [indexed]',
        set_dbname=lambda dbname: state['set_dbname_calls'].append(dbname),
    )
    cli.main_formatter = SimpleNamespace(supported_formats=['ascii', 'csv'])
    cli.completion_refresher = SimpleNamespace(
        refresh=lambda executor, callbacks, options: state['refresh_calls'].append((executor, callbacks, options))
    )
    cli.smart_completion = True
    state['callback'] = callback
    return cli, state


def test_refresh_completions_stops_prefetch() -> None:
    cli, state = make_refresh_cli()

    main.MyCli.refresh_completions(cli)

    assert state['stopped'] == [True]


def test_refresh_completions_returns_started_status() -> None:
    cli, _state = make_refresh_cli()

    result: list[SQLResult] = main.MyCli.refresh_completions(cli)

    assert result == [SQLResult(status='Auto-completion refresh started in the background.')]


def test_refresh_completions_passes_options_to_refresher() -> None:
    cli, state = make_refresh_cli()

    main.MyCli.refresh_completions(cli)

    assert state['refresh_calls'] == [
        (
            cli.sqlexecute,
            state['callback'],
            {
                'smart_completion': True,
                'supported_formats': ['ascii', 'csv'],
                'keyword_casing': 'upper',
                'indexed_column_suffix': ' [indexed]',
            },
        )
    ]


def test_refresh_completions_does_not_update_dbname_without_reset() -> None:
    cli, state = make_refresh_cli()

    main.MyCli.refresh_completions(cli)

    assert state['set_dbname_calls'] == []


def test_refresh_completions_updates_dbname_when_reset() -> None:
    cli = make_bare_mycli()
    set_dbname_calls: list[str] = []
    cli.schema_prefetcher = SimpleNamespace(stop=lambda: None)
    cli.sqlexecute = SimpleNamespace(dbname='next_db')
    cli.completer = SimpleNamespace(
        keyword_casing='lower',
        indexed_column_suffix='*',
        set_dbname=lambda dbname: set_dbname_calls.append(dbname),
    )
    cli.main_formatter = SimpleNamespace(supported_formats=['table'])
    cli.completion_refresher = SimpleNamespace(refresh=lambda executor, callbacks, options: None)

    main.MyCli.refresh_completions(cli, reset=True)

    assert set_dbname_calls == ['next_db']


def test_refresh_completions_uses_lock_when_reset() -> None:
    cli = make_bare_mycli()
    entered_lock = {'count': 0}
    cli.schema_prefetcher = SimpleNamespace(stop=lambda: None)
    cli.sqlexecute = SimpleNamespace(dbname='next_db')
    cli._completer_lock = cast(Any, ReusableLock(lambda: entered_lock.__setitem__('count', entered_lock['count'] + 1)))
    cli.completer = SimpleNamespace(
        keyword_casing='lower',
        indexed_column_suffix='*',
        set_dbname=lambda dbname: None,
    )
    cli.main_formatter = SimpleNamespace(supported_formats=['table'])
    cli.completion_refresher = SimpleNamespace(refresh=lambda executor, callbacks, options: None)

    main.MyCli.refresh_completions(cli, reset=True)

    assert entered_lock == {'count': 1}


def make_refreshed_cli() -> tuple[Any, Any, Any, dict[str, Any]]:
    cli = make_bare_mycli()
    state: dict[str, Any] = {
        'entered_lock': {'count': 0},
        'invalidated': [],
        'prefetch_started': [],
        'copy_calls': [],
    }
    old_completer = SimpleNamespace(dbmetadata={'old': object()})
    new_completer = SimpleNamespace(
        dbname='current',
        copy_other_schemas_from=lambda source, exclude: state['copy_calls'].append((source, exclude)),
    )
    cli._completer_lock = cast(
        Any,
        ReusableLock(lambda: state['entered_lock'].__setitem__('count', state['entered_lock']['count'] + 1)),
    )
    cli.completer = old_completer
    cli.prompt_session = SimpleNamespace(app=SimpleNamespace(invalidate=lambda: state['invalidated'].append(True)))
    cli.schema_prefetcher = SimpleNamespace(start_configured=lambda: state['prefetch_started'].append(True))
    return cli, old_completer, new_completer, state


def test_on_completions_refreshed_swaps_completer() -> None:
    cli, _old_completer, new_completer, _state = make_refreshed_cli()

    main.MyCli._on_completions_refreshed(cli, new_completer)

    assert cli.completer is new_completer


def test_on_completions_refreshed_copies_other_schemas() -> None:
    cli, old_completer, new_completer, state = make_refreshed_cli()

    main.MyCli._on_completions_refreshed(cli, new_completer)

    assert state['copy_calls'] == [(old_completer, 'current')]


def test_on_completions_refreshed_uses_lock() -> None:
    cli, _old_completer, new_completer, state = make_refreshed_cli()

    main.MyCli._on_completions_refreshed(cli, new_completer)

    assert state['entered_lock'] == {'count': 1}


def test_on_completions_refreshed_invalidates_prompt() -> None:
    cli, _old_completer, new_completer, state = make_refreshed_cli()

    main.MyCli._on_completions_refreshed(cli, new_completer)

    assert state['invalidated'] == [True]


def test_on_completions_refreshed_starts_schema_prefetch() -> None:
    cli, _old_completer, new_completer, state = make_refreshed_cli()

    main.MyCli._on_completions_refreshed(cli, new_completer)

    assert state['prefetch_started'] == [True]


def run_query_with_state(monkeypatch, tmp_path, *, warnings_enabled: bool = True) -> dict[str, Any]:
    cli = make_bare_mycli()
    normal_rows = FakeCursorBase(rows=[('one',)], warning_count=1)
    warning_rows = FakeCursorBase(rows=[('warning',)], warning_count=0)
    state: dict[str, Any] = {
        'run_calls': [],
        'logged_queries': [],
        'logged_output': [],
        'formatted': [],
        'echoed': [],
        'checkpoint_path': tmp_path / 'checkpoint.sql',
    }

    def run(sql: str) -> list[SQLResult]:
        state['run_calls'].append(sql)
        if sql == 'SHOW WARNINGS':
            return [SQLResult(rows=warning_rows, status='warnings')]
        return [SQLResult(rows=normal_rows, status='ok')]

    def format_sqlresult(result: SQLResult, **kwargs: Any) -> list[str]:
        state['formatted'].append((result, kwargs))
        return [str(result.status)]

    monkeypatch.setattr(client_query, 'Cursor', FakeCursorBase)
    monkeypatch.setattr(client_query.special, 'is_expanded_output', lambda: True)
    monkeypatch.setattr(client_query.special, 'is_redirected', lambda: False)
    monkeypatch.setattr(client_query.special, 'is_show_warnings_enabled', lambda: warnings_enabled)
    monkeypatch.setattr(client_query.click, 'echo', lambda line, nl=True: state['echoed'].append((line, nl)))

    cli.sqlexecute = SimpleNamespace(run=run)
    cli.log_query = lambda query: state['logged_queries'].append(query)
    cli.log_output = lambda line: state['logged_output'].append(line)
    cli.format_sqlresult = format_sqlresult
    main.MyCli.run_query(cli, 'select 1;\n', checkpoint=str(state['checkpoint_path']), new_line=False)
    state['cli'] = cli
    return state


def test_run_query_executes_query(monkeypatch, tmp_path) -> None:
    state = run_query_with_state(monkeypatch, tmp_path, warnings_enabled=False)

    assert state['run_calls'] == ['select 1;\n']


def test_run_query_logs_query(monkeypatch, tmp_path) -> None:
    state = run_query_with_state(monkeypatch, tmp_path, warnings_enabled=False)

    assert state['logged_queries'] == ['select 1;\n']


def test_run_query_logs_output(monkeypatch, tmp_path) -> None:
    state = run_query_with_state(monkeypatch, tmp_path, warnings_enabled=False)

    assert state['logged_output'] == ['ok']


def test_run_query_echoes_output(monkeypatch, tmp_path) -> None:
    state = run_query_with_state(monkeypatch, tmp_path, warnings_enabled=False)

    assert state['echoed'] == [('ok', False)]


def test_run_query_sets_formatter_query(monkeypatch, tmp_path) -> None:
    state = run_query_with_state(monkeypatch, tmp_path, warnings_enabled=False)
    cli = state['cli']

    assert (cli.main_formatter.query, cli.redirect_formatter.query) == ('select 1;\n', 'select 1;\n')


def test_run_query_passes_display_flags_to_formatter(monkeypatch, tmp_path) -> None:
    state = run_query_with_state(monkeypatch, tmp_path, warnings_enabled=False)

    assert state['formatted'][0][1]['is_expanded'] is True


def test_run_query_uses_redirect_state_for_formatter(monkeypatch, tmp_path) -> None:
    state = run_query_with_state(monkeypatch, tmp_path, warnings_enabled=False)

    assert state['formatted'][0][1]['is_redirected'] is False


def test_run_query_does_not_style_regular_output_as_warnings(monkeypatch, tmp_path) -> None:
    state = run_query_with_state(monkeypatch, tmp_path, warnings_enabled=False)

    assert state['formatted'][0][1].get('is_warnings_style') is not True


def test_run_query_fetches_warnings_when_enabled(monkeypatch, tmp_path) -> None:
    state = run_query_with_state(monkeypatch, tmp_path)

    assert state['run_calls'] == ['select 1;\n', 'SHOW WARNINGS']


def test_run_query_styles_warning_output(monkeypatch, tmp_path) -> None:
    state = run_query_with_state(monkeypatch, tmp_path)

    assert state['formatted'][1][1]['is_warnings_style'] is True


def test_run_query_echoes_warning_output(monkeypatch, tmp_path) -> None:
    state = run_query_with_state(monkeypatch, tmp_path)

    assert state['echoed'][-1] == ('warnings', False)


def test_run_query_writes_checkpoint(monkeypatch, tmp_path) -> None:
    state = run_query_with_state(monkeypatch, tmp_path, warnings_enabled=False)

    assert state['checkpoint_path'].read_text(encoding='utf-8') == 'select 1;\n'


def test_get_last_query_returns_none() -> None:
    cli = make_bare_mycli()

    assert main.MyCli.get_last_query(cli) is None


def test_get_last_query_returns_latest_query() -> None:
    cli = make_bare_mycli()
    cli.query_history = [Query('select 1', True, False), Query('select 2', True, False)]

    assert main.MyCli.get_last_query(cli) == 'select 2'



# type: ignore
"""Regression tests for completing right after a schema (``USE``) switch.

When the user runs ``USE <db>``, ``refresh_completions(reset=True)`` updates the
live completer's ``dbname`` *immediately* so unqualified completions reflect the
switch, while the matching table metadata is still being fetched on a background
thread (and only swapped in when the refresh finishes).

In that window ``self.dbname`` names a schema that is not yet a key in
``dbmetadata["tables"]``.  A naked ``SELECT `` completion reaches
``populate_scoped_cols`` with no scoped tables and used to crash with
``KeyError`` on ``meta["tables"][self.dbname]``.  These tests pin the safe
behaviour: suggest ``*`` during the window, real columns once loaded.
"""

import threading
import traceback

from prompt_toolkit.document import Document

from mycli.sqlcompleter import SQLCompleter


def _make_completer() -> SQLCompleter:
    completer = SQLCompleter(smart_completion=True)
    completer.load_schema_metadata(
        schema="old_db",
        table_columns={"orders": ["*", "id", "total"]},
        indexed_columns={"orders": {"id"}},
        foreign_keys={},
        enum_values={},
        functions={},
        procedures={},
    )
    completer.set_dbname("old_db")
    return completer


def test_populate_scoped_cols_unloaded_schema_returns_star() -> None:
    completer = _make_completer()
    # Switch to a schema whose metadata has not been loaded yet.
    completer.set_dbname("new_db")
    assert "new_db" not in completer.dbmetadata["tables"]

    # Empty scoped tables => the "naked SELECT" branch that does the lookup.
    assert completer.populate_scoped_cols([]) == ["*"]


def test_get_completions_after_use_switch_before_refresh() -> None:
    completer = _make_completer()
    completer.set_dbname("new_db")  # metadata not loaded yet

    for text in ("SELECT ", "SELECT col", "SELECT a, "):
        document = Document(text, len(text))
        # Must not raise KeyError while the background refresh is in flight.
        completions = list(completer.get_completions(document, None))
        assert all(c.text for c in completions)


def test_columns_available_once_schema_loads() -> None:
    completer = _make_completer()
    completer.set_dbname("new_db")
    # Background refresh finishes and loads the new schema.
    completer.load_schema_metadata(
        schema="new_db",
        table_columns={"customers": ["*", "name", "email"]},
        indexed_columns={"customers": {"email"}},
        foreign_keys={},
        enum_values={},
        functions={},
        procedures={},
    )

    cols = completer.populate_scoped_cols([])
    assert "name" in cols
    assert "email" in cols


def test_completion_during_concurrent_use_switch_does_not_crash() -> None:
    """A reader must survive a writer flipping ``dbname`` between schemas.

    Mirrors the live REPL: prompt_toolkit's completion thread reads the
    completer lock-free while a background refresh switches ``dbname`` and
    loads/unloads schema metadata.
    """
    completer = _make_completer()
    stop = threading.Event()
    errors: list[str] = []

    def reader() -> None:
        document = Document("SELECT ", len("SELECT "))
        while not stop.is_set():
            try:
                list(completer.get_completions(document, None))
            except Exception:
                errors.append(traceback.format_exc())
                return

    def writer() -> None:
        n = 0
        while not stop.is_set() and n < 2000:
            n += 1
            schema = f"db_{n}"
            # Point dbname at a not-yet-loaded schema, then load it, as the
            # reset=True refresh path does.
            completer.set_dbname(schema)
            completer.load_schema_metadata(
                schema=schema,
                table_columns={"t": ["*", "c1", "c2"]},
                indexed_columns={"t": {"c1"}},
                foreign_keys={},
                enum_values={},
                functions={},
                procedures={},
            )

    threads = [threading.Thread(target=reader) for _ in range(3)]
    threads.append(threading.Thread(target=writer))
    for thread in threads:
        thread.start()
    threads[-1].join(timeout=5)
    stop.set()
    for thread in threads:
        thread.join(timeout=5)

    assert not errors, "completion crashed during USE switch:\n" + "\n".join(errors)



from __future__ import annotations

import os
import re
import sys
from typing import TYPE_CHECKING, Any, Callable
from urllib.parse import parse_qs, unquote, urlparse

import click

from mycli.config import str_to_bool
from mycli.constants import EMPTY_PASSWORD_FLAG_SENTINEL
from mycli.main_modes.batch import main_batch_from_stdin, main_batch_with_progress_bar, main_batch_without_progress_bar
from mycli.main_modes.checkup import main_checkup
from mycli.main_modes.execute import main_execute_from_cli
from mycli.main_modes.list_dsn import main_list_dsn
from mycli.packages.cli_utils import is_valid_connection_scheme
from mycli.password_sources import PasswordCandidates
from mycli.vault import (
    DEFAULT_VAULT_EXECUTABLE,
    DEFAULT_VAULT_PASSWORD_FIELD,
    DEFAULT_VAULT_USERNAME_FIELD,
    VaultError,
    get_field_from_vault,
)

if TYPE_CHECKING:
    from mycli.main import CliArgs

ClientFactory = Callable[..., Any]
ENV_VAR_PATTERN = re.compile(r'^\$\{([A-Za-z_][A-Za-z0-9_]*)\}$')
KNOWN_DSN_QUERY_PARAMS = {
    'character_set',
    'keepalive_ticks',
    'socket',
    'ssh_jump',
    'ssl_ca',
    'ssl_capath',
    'ssl_cert',
    'ssl_cipher',
    'ssl_key',
    'ssl_mode',
    'ssl_verify_server_cert',
    'tls_version',
    'vault_address',
    'vault_mount',
    'vault_secret',
    'vault_password_field',
    'vault_username_field',
}


class DsnAliasEnvVarError(ValueError):
    pass


def expand_dsn_alias_env_var(value: str | None, alias_name: str) -> str | None:
    if value is None:
        return None

    match = ENV_VAR_PATTERN.fullmatch(value)
    if not match:
        return value

    var_name = match.group(1)
    try:
        return os.environ[var_name]
    except KeyError as exc:
        raise DsnAliasEnvVarError(f'Environment variable {var_name} referenced by DSN alias {alias_name} is not set.') from exc


def split_dsn_netloc(netloc: str) -> tuple[str | None, str | None, str | None, str | None]:
    username = None
    password = None
    host_port = netloc

    if '@' in host_port:
        user_info, host_port = host_port.rsplit('@', 1)
        username, separator, password = user_info.partition(':')
        if not separator:
            password = None

    if not host_port:
        return username, password, None, None

    if host_port.startswith('['):
        end = host_port.find(']')
        if end >= 0:
            host = host_port[1:end]
            port = host_port[end + 2 :] if host_port[end + 1 : end + 2] == ':' else None
            return username, password, host, port

    host, separator, port = host_port.partition(':')
    return username, password, host or None, port if separator else None


def expand_dsn_alias_env_vars(
    dsn_uri: str, alias_name: str
) -> tuple[str | None, str | None, str | None, int | None, str, dict[str, list[str]]]:
    uri = urlparse(dsn_uri)
    username, password, host, port = split_dsn_netloc(uri.netloc)

    expanded_port = expand_dsn_alias_env_var(port, alias_name)
    try:
        port_number = int(expanded_port) if expanded_port else None
    except ValueError as exc:
        raise DsnAliasEnvVarError(f'Port in DSN alias {alias_name} must be an integer.') from exc

    params = {
        key: [expand_dsn_alias_env_var(value, alias_name) or '' for value in values]
        for key, values in parse_qs(uri.query, keep_blank_values=True).items()
    }

    return (
        expand_dsn_alias_env_var(unquote(username) if username is not None else None, alias_name),
        expand_dsn_alias_env_var(unquote(password) if password is not None else None, alias_name),
        expand_dsn_alias_env_var(host, alias_name),
        port_number,
        expand_dsn_alias_env_var(uri.path[1:], alias_name) or '',
        params,
    )


def run_from_cli_args(cli_args: 'CliArgs', client_factory: ClientFactory) -> None:
    from mycli import main as main_module

    cli_verbosity = main_module.preprocess_cli_args(cli_args, is_valid_connection_scheme)

    mycli = client_factory(
        prompt=cli_args.prompt,
        toolbar_format=cli_args.toolbar,
        logfile=cli_args.logfile,
        login_path=cli_args.login_path,
        auto_vertical_output=cli_args.auto_vertical_output,
        warn=cli_args.warn,
        myclirc=cli_args.myclirc,
        show_warnings=cli_args.show_warnings,
        cli_verbosity=cli_verbosity,
    )

    if cli_args.checkup:
        main_checkup(mycli)
        sys.exit(0)

    if cli_args.csv and cli_args.format not in [None, 'csv']:
        click.secho("Conflicting --csv and --format arguments.", err=True, fg="red")
        sys.exit(1)

    if cli_args.table and cli_args.format not in [None, 'table']:
        click.secho("Conflicting --table and --format arguments.", err=True, fg="red")
        sys.exit(1)

    if not cli_args.format:
        cli_args.format = 'default'

    if cli_args.csv:
        cli_args.format = 'csv'

    if cli_args.table:
        cli_args.format = 'table'

    if cli_args.list_dsn:
        sys.exit(main_list_dsn(mycli))

    # Choose which ever one has a valid value.
    database = cli_args.dbname or cli_args.database

    dsn_uri = None
    dsn_password: str | None = None

    # Treat the database argument as a DSN alias only if it matches a configured alias
    # todo why is port tested but not socket?
    truthy_password = (
        cli_args.password not in (None, EMPTY_PASSWORD_FLAG_SENTINEL)
        or cli_args.password_file is not None
        or os.environ.get('MYSQL_PWD') is not None
    )
    if (
        database
        and "://" not in database
        and not any([
            cli_args.user,
            truthy_password,
            cli_args.host,
            cli_args.port,
            cli_args.login_path,
        ])
        and database in mycli.config.get("alias_dsn", {})
    ):
        cli_args.dsn, database = database, ""

    if database and "://" in database:
        dsn_uri, database = database, ""

    if cli_args.dsn:
        try:
            dsn_uri = mycli.config["alias_dsn"][cli_args.dsn]
        except KeyError:
            is_valid_scheme, scheme = is_valid_connection_scheme(cli_args.dsn)
            if is_valid_scheme:
                dsn_uri = cli_args.dsn
            else:
                click.secho(
                    "Could not find the specified DSN in the config file. Please check the \"[alias_dsn]\" section in your myclirc.",
                    err=True,
                    fg="red",
                )
                sys.exit(1)
        else:
            mycli.dsn_alias = cli_args.dsn

    if dsn_uri:
        is_valid_scheme, scheme = is_valid_connection_scheme(dsn_uri)
        if not is_valid_scheme:
            click.secho(f'Error: Unknown connection scheme provided for DSN URI ({scheme}://)', err=True, fg='red')
            sys.exit(1)

        uri = urlparse(dsn_uri)
        env_var_alias_name = None
        dsn_alias = getattr(mycli, 'dsn_alias', None)
        if dsn_alias and str_to_bool(mycli.config['main'].get('expand_dsn_alias_env_vars', 'False')):
            env_var_alias_name = dsn_alias

        if env_var_alias_name:
            try:
                dsn_user, dsn_password, dsn_host, dsn_port, dsn_database, dsn_params = expand_dsn_alias_env_vars(
                    dsn_uri, env_var_alias_name
                )
            except DsnAliasEnvVarError as exc:
                click.secho(str(exc), err=True, fg='red')
                sys.exit(1)
        else:
            dsn_user = unquote(uri.username) if uri.username is not None else None
            dsn_password = unquote(uri.password) if uri.password is not None else None
            dsn_host = uri.hostname
            dsn_port = uri.port
            dsn_database = uri.path[1:]
            dsn_params = parse_qs(uri.query, keep_blank_values=True) if uri.query else {}

        if not database:
            database = dsn_database
        if not cli_args.user and dsn_user is not None:
            cli_args.user = dsn_user
        if not cli_args.host:
            cli_args.host = dsn_host
        if not cli_args.port:
            cli_args.port = dsn_port

        if unknown_dsn_params := sorted(set(dsn_params) - KNOWN_DSN_QUERY_PARAMS):
            click.secho(
                f'Warning: Ignored unknown DSN URI query parameters: {", ".join(unknown_dsn_params)}.',
                err=True,
                fg='yellow',
            )

        if params := dsn_params.get('ssl_mode'):
            cli_args.ssl_mode = cli_args.ssl_mode or params[0]
        if params := dsn_params.get('ssl_ca'):
            cli_args.ssl_ca = cli_args.ssl_ca or params[0]
            cli_args.ssl_mode = cli_args.ssl_mode or 'on'
        if params := dsn_params.get('ssl_capath'):
            cli_args.ssl_capath = cli_args.ssl_capath or params[0]
            cli_args.ssl_mode = cli_args.ssl_mode or 'on'
        if params := dsn_params.get('ssl_cert'):
            cli_args.ssl_cert = cli_args.ssl_cert or params[0]
            cli_args.ssl_mode = cli_args.ssl_mode or 'on'
        if params := dsn_params.get('ssl_key'):
            cli_args.ssl_key = cli_args.ssl_key or params[0]
            cli_args.ssl_mode = cli_args.ssl_mode or 'on'
        if params := dsn_params.get('ssl_cipher'):
            cli_args.ssl_cipher = cli_args.ssl_cipher or params[0]
            cli_args.ssl_mode = cli_args.ssl_mode or 'on'
        if params := dsn_params.get('tls_version'):
            cli_args.tls_version = cli_args.tls_version or params[0]
            cli_args.ssl_mode = cli_args.ssl_mode or 'on'
        if params := dsn_params.get('ssl_verify_server_cert'):
            cli_args.ssl_verify_server_cert = cli_args.ssl_verify_server_cert or (params[0].lower() == 'true')
            cli_args.ssl_mode = cli_args.ssl_mode or 'on'
        if params := dsn_params.get('socket'):
            cli_args.socket = cli_args.socket or params[0]
        if params := dsn_params.get('keepalive_ticks'):
            if cli_args.keepalive_ticks is None:
                cli_args.keepalive_ticks = int(params[0])
        if params := dsn_params.get('character_set'):
            cli_args.character_set = cli_args.character_set or params[0]
        if params := dsn_params.get('ssh_jump'):
            cli_args.ssh_jump = cli_args.ssh_jump or params[0]
        if params := dsn_params.get('vault_address'):
            cli_args.vault_address = cli_args.vault_address or params[0]
        if params := dsn_params.get('vault_mount'):
            cli_args.vault_mount = cli_args.vault_mount or params[0]
        if params := dsn_params.get('vault_secret'):
            cli_args.vault_secret = cli_args.vault_secret or params[0]
        if params := dsn_params.get('vault_password_field'):
            cli_args.vault_password_field = cli_args.vault_password_field or params[0]
        if params := dsn_params.get('vault_username_field'):
            cli_args.vault_username_field = cli_args.vault_username_field or params[0]

    keepalive_ticks = cli_args.keepalive_ticks if cli_args.keepalive_ticks is not None else mycli.default_keepalive_ticks
    ssl_mode = cli_args.ssl_mode or mycli.ssl_mode

    if ssl_mode in ("auto", "on"):
        if cli_args.socket and ssl_mode == 'auto':
            ssl = None
        else:
            ssl = {
                "mode": ssl_mode,
                "ca": cli_args.ssl_ca and os.path.expanduser(cli_args.ssl_ca),
                "cert": cli_args.ssl_cert and os.path.expanduser(cli_args.ssl_cert),
                "key": cli_args.ssl_key and os.path.expanduser(cli_args.ssl_key),
                "capath": cli_args.ssl_capath,
                "cipher": cli_args.ssl_cipher,
                "tls_version": cli_args.tls_version,
                "check_hostname": cli_args.ssl_verify_server_cert,
            }
            # remove empty ssl options
            ssl = {k: v for k, v in ssl.items() if v is not None}
    else:
        ssl = None

    # Merge init-commands: global, DSN-specific, then CLI
    init_cmds: list[str] = []
    # 1) Global init-commands
    global_section = mycli.config.get("init-commands", {})
    for _, val in global_section.items():
        if isinstance(val, (list, tuple)):
            init_cmds.extend(val)
        elif val:
            init_cmds.append(val)
    # 2) DSN-specific init-commands
    if cli_args.dsn:
        alias_section = mycli.config.get("alias_dsn.init-commands", {})
        if cli_args.dsn in alias_section:
            val = alias_section.get(cli_args.dsn)
            if isinstance(val, (list, tuple)):
                init_cmds.extend(val)
            elif val:
                init_cmds.append(val)
    # 3) CLI-provided init_command
    if cli_args.init_command:
        init_cmds.append(cli_args.init_command)

    combined_init_cmd = "; ".join(cmd.strip() for cmd in init_cmds if cmd)

    if cli_args.use_keyring is not None and cli_args.use_keyring.lower() == 'reset':
        use_keyring = True
        reset_keyring = True
    elif cli_args.use_keyring is not None and cli_args.use_keyring.lower() == 'auto':
        if os.environ.get('SSH_CONNECTION'):
            use_keyring = False
            reset_keyring = False
        else:
            use_keyring = True
            reset_keyring = False
    elif cli_args.use_keyring is None:
        if mycli.config['main'].get('use_keyring', 'False').lower() == 'auto':
            if os.environ.get('SSH_CONNECTION'):
                use_keyring = False
                reset_keyring = False
            else:
                use_keyring = True
                reset_keyring = False
        else:
            use_keyring = str_to_bool(mycli.config['main'].get('use_keyring', 'False'))
            reset_keyring = False
    else:
        use_keyring = str_to_bool(cli_args.use_keyring)
        reset_keyring = False

    password_candidates = PasswordCandidates()
    if cli_args.password == EMPTY_PASSWORD_FLAG_SENTINEL:
        password_candidates.add_value('prompt', cli_args.password)
    elif cli_args.password is not None:
        password_candidates.add_value('literal', cli_args.password)
    if cli_args.password_file:
        password_candidates.add_loader('file', lambda: main_module.get_password_from_file(cli_args.password_file))
    if os.environ.get('MYSQL_PWD') is not None:
        password_candidates.add_value('environment', os.environ.get('MYSQL_PWD'))
    if dsn_password is not None:
        password_candidates.add_value('dsn', dsn_password)

    if cli_args.vault_secret:
        vault_secret = cli_args.vault_secret
        vault_config = mycli.config.get('vault_beta', {})
        vault_address = cli_args.vault_address or os.environ.get('VAULT_ADDR') or vault_config.get('address') or None
        vault_mount = cli_args.vault_mount or vault_config.get('default_mount') or None
        vault_password_field = cli_args.vault_password_field or vault_config.get('default_password_field') or DEFAULT_VAULT_PASSWORD_FIELD
        vault_executable = vault_config.get('vault_executable') or DEFAULT_VAULT_EXECUTABLE

        def load_vault_password() -> str | None:
            try:
                return get_field_from_vault(
                    vault_password_field,
                    vault_secret,
                    executable=vault_executable,
                    mount=vault_mount,
                    address=vault_address,
                )
            except VaultError as exc:
                click.secho(f'Error reading password from Vault: {exc}', err=True, fg='red')
                sys.exit(1)

        password_candidates.add_loader('vault', load_vault_password)

    if cli_args.user is None and cli_args.vault_secret:
        vault_config = mycli.config.get('vault_beta', {})
        vault_address = cli_args.vault_address or os.environ.get('VAULT_ADDR') or vault_config.get('address') or None
        vault_mount = cli_args.vault_mount or vault_config.get('default_mount') or None
        vault_username_field = cli_args.vault_username_field or vault_config.get('default_username_field') or DEFAULT_VAULT_USERNAME_FIELD
        vault_executable = vault_config.get('vault_executable') or DEFAULT_VAULT_EXECUTABLE
        try:
            vault_username = get_field_from_vault(
                vault_username_field,
                cli_args.vault_secret,
                executable=vault_executable,
                mount=vault_mount,
                address=vault_address,
            )
        except VaultError as exc:
            click.secho(f'Error reading username from Vault: {exc}', err=True, fg='red')
            sys.exit(1)
    else:
        vault_username = None
    vault_username_from_vault = vault_username is not None

    try:
        mycli.connect(
            database=database,
            user=vault_username if cli_args.user is None else cli_args.user,
            password_candidates=password_candidates,
            host=cli_args.host,
            port=cli_args.port,
            socket=cli_args.socket,
            local_infile=cli_args.local_infile,
            ssl=ssl,
            init_command=combined_init_cmd,
            unbuffered=cli_args.unbuffered,
            character_set=cli_args.character_set,
            use_keyring=use_keyring,
            reset_keyring=reset_keyring,
            keepalive_ticks=keepalive_ticks,
            ssh_jump=cli_args.ssh_jump,
            ssh_cli_options=cli_args.ssh_options,
            vault_address=cli_args.vault_address,
            vault_mount=cli_args.vault_mount,
            vault_secret=cli_args.vault_secret,
            vault_password_field=cli_args.vault_password_field,
            vault_username_field=cli_args.vault_username_field,
            vault_username_from_vault=vault_username_from_vault,
        )

        if combined_init_cmd:
            click.echo(f"Executing init-command: {combined_init_cmd}", err=True)

        mycli.logger.debug(
            "Launch Params: \n\tdatabase: %r\tuser: %r\thost: %r\tport: %r",
            database,
            cli_args.user,
            cli_args.host,
            cli_args.port,
        )

        if cli_args.execute is not None:
            sys.exit(main_execute_from_cli(mycli, cli_args))

        if cli_args.batch is not None and cli_args.batch != '-' and cli_args.progress and sys.stderr.isatty():
            sys.exit(main_batch_with_progress_bar(mycli, cli_args))

        if cli_args.batch is not None:
            sys.exit(main_batch_without_progress_bar(mycli, cli_args))

        if not sys.stdin.isatty():
            sys.exit(main_batch_from_stdin(mycli, cli_args))

        mycli.run_cli()
    finally:
        mycli.close()



from __future__ import annotations

from io import TextIOWrapper
import logging
import os
import threading
from typing import IO, Literal

from cli_helpers.tabular_output import TabularOutputFormatter
from configobj import ConfigObj
from prompt_toolkit.formatted_text import to_formatted_text
from prompt_toolkit.shortcuts import PromptSession
import sqlparse

from mycli.app_state import (
    AppStateMixin,
    configure_prompt_state,
    destructive_keywords_from_config,
    llm_prompt_truncation,
    normalize_image_protocol,
    normalize_ssl_mode,
)
from mycli.client_commands import ClientCommandsMixin
from mycli.client_connection import ClientConnectionMixin
from mycli.client_query import ClientQueryMixin
from mycli.clistyle import style_factory_helpers, style_factory_ptoolkit
from mycli.completion_refresher import CompletionRefresher
from mycli.config import (
    get_mylogin_cnf_path,
    open_mylogin_cnf,
    read_config_file,
    read_config_files,
    write_default_config,
)
from mycli.constants import DEFAULT_PROMPT
from mycli.main_modes import repl as repl_package
from mycli.output import OutputMixin
from mycli.packages import special
from mycli.packages.special.dsn_aliases import DsnAliases
from mycli.packages.special.favoritequeries import FavoriteQueries
from mycli.packages.tabular_output import sql_format
from mycli.schema_prefetcher import SchemaPrefetcher
from mycli.sqlcompleter import SQLCompleter
from mycli.sqlexecute import SQLExecute
from mycli.ssh_tunnel import SshTunnel
from mycli.types import Query

sqlparse.engine.grouping.MAX_GROUPING_DEPTH = None  # type: ignore[assignment]
sqlparse.engine.grouping.MAX_GROUPING_TOKENS = None  # type: ignore[assignment]


class MyCli(AppStateMixin, OutputMixin, ClientCommandsMixin, ClientConnectionMixin, ClientQueryMixin):
    default_prompt = DEFAULT_PROMPT
    default_prompt_splitln = "\\u@\\h\\n(\\t):\\d>"
    max_len_prompt = 45
    prompt_lines: int
    sqlexecute: SQLExecute | None
    numeric_alignment: str

    # check XDG_CONFIG_HOME exists and not an empty string
    xdg_config_home = os.environ.get("XDG_CONFIG_HOME", "~/.config")
    xdg_config_file = os.path.join(os.path.expanduser(xdg_config_home), "mycli", "myclirc")
    system_config_files: list[str | IO[str]] = [
        "/etc/myclirc",
    ]

    def __init__(
        self,
        sqlexecute: SQLExecute | None = None,
        prompt: str | None = None,
        toolbar_format: str | None = None,
        logfile: TextIOWrapper | Literal[False] | None = None,
        login_path: str | None = None,
        auto_vertical_output: bool = False,
        warn: bool | None = None,
        myclirc: str | None = None,
        show_warnings: bool | None = None,
        cli_verbosity: int = 0,
    ) -> None:
        self.sqlexecute = sqlexecute
        self.ssh_tunnel: SshTunnel | None = None
        self.logfile = logfile
        self.login_path = login_path
        self.toolbar_error_message: str | None = None
        self.prompt_session: PromptSession | None = None
        self._keepalive_counter = 0
        self.keepalive_ticks: int | None = 0
        self.sandbox_mode: bool = False
        self.checkpoint: IO | None = None

        if myclirc is None:
            if os.path.exists(self.xdg_config_file):
                myclirc = self.xdg_config_file
            else:
                myclirc = '~/.myclirc'

        # Load config.
        config_files: list[str | IO[str]] = self.system_config_files + [myclirc]

        c = self.config = read_config_files(config_files)
        # only needed in --checkup mode. todo: only load when needed
        self.config_without_package_defaults = read_config_files(config_files, ignore_package_defaults=True)
        # only needed in --checkup mode. todo: only load when needed
        self.config_without_user_options = read_config_files(config_files, ignore_user_options=True)
        self.multi_line = c["main"].as_bool("multi_line")
        self.key_bindings = c["main"]["key_bindings"]
        self.emacs_ttimeoutlen = c['keys'].as_float('emacs_ttimeoutlen')
        self.vi_ttimeoutlen = c['keys'].as_float('vi_ttimeoutlen')
        special.set_timing_enabled(c["main"].as_bool("timing"))
        special.set_show_favorite_query(c["main"].as_bool("show_favorite_query"))
        if show_warnings is not None:
            special.set_show_warnings_enabled(show_warnings)
        else:
            special.set_show_warnings_enabled(c['main'].as_bool('show_warnings'))
        self.beep_after_seconds = float(c["main"]["beep_after_seconds"] or 0)
        self.default_keepalive_ticks = c['connection'].as_int('default_keepalive_ticks')

        FavoriteQueries.instance = FavoriteQueries.from_config(self.config)
        DsnAliases.instance = DsnAliases.from_config(self.config)

        self.dsn_alias: str | None = None
        self.main_formatter = TabularOutputFormatter(format_name=c["main"]["table_format"])
        self.redirect_formatter = TabularOutputFormatter(format_name=c["main"].get("redirect_format", "csv"))
        self.explorer_formatter = TabularOutputFormatter(format_name=c['explorer'].get('format', self.main_formatter))
        sql_format.register_new_formatter(self.main_formatter)
        sql_format.register_new_formatter(self.redirect_formatter)
        sql_format.register_new_formatter(self.explorer_formatter)
        self.main_formatter.mycli = self
        self.redirect_formatter.mycli = self
        self.explorer_formatter.mycli = self
        self.explorer_command = c['explorer'].get('explorer_command')
        self.explorer_trim_footer = c['explorer'].as_bool('trim_footer')
        self.syntax_style = c["main"]["syntax_style"]
        self.verbosity = -1 if c["main"].as_bool("less_chatty") else 0
        if cli_verbosity:
            self.verbosity = cli_verbosity
        self.cli_style = c["colors"]
        self.ptoolkit_style = style_factory_ptoolkit(self.syntax_style, self.cli_style)
        self.helpers_style = style_factory_helpers(self.syntax_style, self.cli_style)
        self.helpers_warnings_style = style_factory_helpers(self.syntax_style, self.cli_style, warnings=True)
        self.wider_completion_menu = c["main"].as_bool("wider_completion_menu")
        c_dest_warning = c["main"].as_bool("destructive_warning")
        self.destructive_warning = c_dest_warning if warn is None else warn
        self.login_path_as_host = c["main"].as_bool("login_path_as_host")
        self.post_redirect_command = c['main'].get('post_redirect_command')
        self.null_string = c['main'].get('null_string')
        self.numeric_alignment = c['main'].get('numeric_alignment', 'right') or 'right'
        self.binary_display = c['main'].get('binary_display')
        self.image_protocol, image_protocol_error = normalize_image_protocol(c['dataframe'].get('image_protocol'))
        if image_protocol_error:
            self.echo(image_protocol_error, err=True, fg='red')
        self.plot_scale_factor = c['dataframe'].as_float('plot_scale_factor')
        self.plot_ppi = c['dataframe'].as_int('plot_ppi')
        self.plot_theme = c['dataframe'].get('plot_theme', 'carbong90') or 'carbong90'
        self.llm_prompt_field_truncate, self.llm_prompt_section_truncate = llm_prompt_truncation(c)

        self.ssl_mode, ssl_mode_error = normalize_ssl_mode(c, self.config_without_package_defaults)
        if ssl_mode_error:
            self.echo(ssl_mode_error, err=True, fg="red")

        # read from cli argument or user config file
        self.auto_vertical_output = auto_vertical_output or c["main"].as_bool("auto_vertical_output")

        # Write user config if system config wasn't the last config loaded.
        if c.filename not in self.system_config_files and not os.path.exists(myclirc):
            write_default_config(myclirc)

        # audit log
        if self.logfile is None and "audit_log" in c["main"]:
            try:
                self.logfile = open(os.path.expanduser(c["main"]["audit_log"]), "a")
            except (IOError, OSError):
                self.echo("Error: Unable to open the audit log file. Your queries will not be logged.", err=True, fg="red")
                self.logfile = False

        self.completion_refresher = CompletionRefresher(self._invalidate_prompt_session)
        self.prefetch_schemas_mode = c["main"].get("prefetch_schemas_mode", "always") or "always"
        raw_prefetch_list = c["main"].as_list("prefetch_schemas_list") if "prefetch_schemas_list" in c["main"] else []
        self.prefetch_schemas_list = [s.strip() for s in raw_prefetch_list if s and s.strip()]
        self.schema_prefetcher = SchemaPrefetcher(self)

        self.logger = logging.getLogger(__name__)
        self.initialize_logging()

        keyword_casing = c["main"].get("keyword_casing", "auto")
        indexed_column_suffix = c['main'].get('indexed_column_suffix', '')

        self.highlight_preview = c['search'].as_bool('highlight_preview')

        self.query_history: list[Query] = []

        # Initialize completer.
        self.smart_completion = c["main"].as_bool("smart_completion")
        self.completer = SQLCompleter(
            self.smart_completion,
            supported_formats=self.main_formatter.supported_formats,
            keyword_casing=keyword_casing,
            indexed_column_suffix=indexed_column_suffix,
        )
        self._completer_lock = threading.Lock()

        self.min_completion_trigger = c["main"].as_int("min_completion_trigger")
        # a hack, pending a better way to handle settings and state
        repl_package.MIN_COMPLETION_TRIGGER = self.min_completion_trigger
        self.last_prompt_message = to_formatted_text('')
        self.last_custom_toolbar_message = to_formatted_text('')

        # Register custom special commands.
        self.register_special_commands()

        # Load .mylogin.cnf if it exists.
        self.mylogin_cnf = ConfigObj()
        mylogin_cnf_h = None
        if mylogin_cnf_path := get_mylogin_cnf_path():
            mylogin_cnf_h = open_mylogin_cnf(mylogin_cnf_path)
            if mylogin_cnf_h:
                self.mylogin_cnf = read_config_file(mylogin_cnf_h, list_values=False) or ConfigObj()
            else:
                print("Error: Unable to read login path file.")

        configure_prompt_state(self, c, prompt, toolbar_format)
        self.prompt_session = None
        self.destructive_keywords = destructive_keywords_from_config(c)
        special.set_destructive_keywords(self.destructive_keywords)

    def _invalidate_prompt_session(self) -> None:
        if self.prompt_session:
            self.prompt_session.app.invalidate()

    def close(self) -> None:
        try:
            self.schema_prefetcher.stop()
        except Exception:
            pass
        if self.sqlexecute is not None:
            try:
                self.sqlexecute.close()
            except Exception:
                pass
        if self.ssh_tunnel is not None:
            try:
                self.ssh_tunnel.close()
            except Exception:
                pass

    def run_cli(self) -> None:
        repl_package.main_repl(self)



# type: ignore

from behave import then, when
import wrappers

from test.features.environment import get_db_name_from_context
from test.features.steps.utils import parse_cli_args_to_dict

TEST_LOGIN_PATH = "test_login_path"


@when('we run mycli with arguments "{exact_args}" without arguments "{excluded_args}"')
@when('we run mycli without arguments "{excluded_args}"')
def step_run_cli_without_args(context, excluded_args, exact_args=""):
    wrappers.run_cli(context, run_args=parse_cli_args_to_dict(exact_args), exclude_args=parse_cli_args_to_dict(excluded_args).keys())


@then('status contains "{expression}"')
def status_contains(context, expression):
    wrappers.expect_exact(context, f"{expression}", timeout=5)

    # Normally, the shutdown after scenario waits for the prompt.
    # But we may have changed the prompt, depending on parameters,
    # so let's wait for its last character
    context.cli.expect_exact(">")
    context.atprompt = True


@then("we are logged in")
def we_are_logged_in(context):
    db_name = get_db_name_from_context(context)
    context.cli.expect_exact(f"{db_name}>", timeout=5)
    context.atprompt = True



from __future__ import annotations

from configobj import ConfigObj
import pytest

from mycli.app_state import (
    AppStateMixin,
    configure_prompt_state,
    destructive_keywords_from_config,
    llm_prompt_truncation,
    normalize_image_protocol,
    normalize_ssl_mode,
)


class AppState(AppStateMixin):
    def __init__(self, login_path: str | None = None) -> None:
        self.login_path = login_path


class PromptState:
    prompt_format: str
    prompt_lines: int
    multiline_continuation_char: str
    toolbar_format: str
    terminal_tab_title_format: str
    terminal_window_title_format: str
    multiplex_window_title_format: str
    multiplex_pane_title_format: str

    def __init__(self) -> None:
        self.default_prompt = 'default> '


@pytest.mark.parametrize(
    ('image_protocol', 'expected'),
    [
        ('iterm2', ('iterm2', None)),
        ('kitty', ('kitty', None)),
        ('none', ('none', None)),
        ('', ('none', None)),
        (None, ('none', None)),
        ('unknown', ('none', 'Invalid config option provided for image_protocol (unknown); disabling.')),
    ],
)
def test_normalize_image_protocol(image_protocol: str | None, expected: tuple[str, str | None]) -> None:
    assert normalize_image_protocol(image_protocol) == expected


@pytest.mark.parametrize(
    ('prompt', 'config_prompt', 'toolbar_format', 'config_toolbar', 'expected_prompt', 'expected_toolbar'),
    [
        ('custom> ', 'configured> ', 'custom toolbar', 'configured toolbar', 'custom> ', 'custom toolbar'),
        (None, 'configured> ', None, 'configured toolbar', 'configured> ', 'configured toolbar'),
        ('', '', '', 'configured toolbar', 'default> ', 'configured toolbar'),
    ],
)
def test_configure_prompt_state_uses_overrides_and_fallbacks(
    prompt: str | None,
    config_prompt: str,
    toolbar_format: str | None,
    config_toolbar: str,
    expected_prompt: str,
    expected_toolbar: str,
) -> None:
    state = PromptState()
    config = ConfigObj({
        'main': {
            'prompt': config_prompt,
            'prompt_continuation': '... ',
            'toolbar': config_toolbar,
            'terminal_tab_title': 'tab title',
            'terminal_window_title': 'window title',
            'multiplex_window_title': 'multiplex window title',
            'multiplex_pane_title': 'multiplex pane title',
        },
    })

    configure_prompt_state(state, config, prompt, toolbar_format)  # type: ignore[arg-type]

    assert state.prompt_format == expected_prompt
    assert state.prompt_lines == 0
    assert state.multiline_continuation_char == '... '
    assert state.toolbar_format == expected_toolbar
    assert state.terminal_tab_title_format == 'tab title'
    assert state.terminal_window_title_format == 'window title'
    assert state.multiplex_window_title_format == 'multiplex window title'
    assert state.multiplex_pane_title_format == 'multiplex pane title'


@pytest.mark.parametrize('ssl_mode', ['auto', 'on', 'off'])
def test_normalize_ssl_mode_accepts_known_values(ssl_mode: str) -> None:
    config = ConfigObj({'main': {'ssl_mode': ssl_mode}, 'connection': {'default_ssl_mode': ssl_mode}})
    config_wo = ConfigObj({'main': {}, 'connection': {}})

    assert normalize_ssl_mode(config, config_wo) == (ssl_mode, None)


def test_normalize_ssl_mode_falls_back_to_connection_default() -> None:
    config = ConfigObj({'main': {'ssl_mode': ''}, 'connection': {'default_ssl_mode': 'on'}})
    config_wo = ConfigObj({'main': {}, 'connection': {}})

    assert normalize_ssl_mode(config, config_wo) == ('on', None)


def test_normalize_ssl_mode_returns_none_when_not_configured() -> None:
    config = ConfigObj({'main': {}, 'connection': {}})
    config_wo = ConfigObj({'main': {}, 'connection': {}})

    assert normalize_ssl_mode(config, config_wo) == (None, None)


def test_normalize_ssl_mode_migrates_deprecated_main_value() -> None:
    config = ConfigObj({'main': {}, 'connection': {'default_ssl_mode': 'off'}})
    config_wo = ConfigObj({'main': {'ssl_mode': 'on'}})

    ssl_mode, warning = normalize_ssl_mode(config, config_wo)

    assert ssl_mode == 'on'
    assert (
        warning == 'Mycli 2.0 migration: automatically moving ssl_mode under [main] to default_ssl_mode under [connection] in ~/.myclirc .'
    )
    assert config_wo['connection']['default_ssl_mode'] == 'on'
    assert 'ssl_mode' not in config_wo['main']


def test_normalize_ssl_mode_uses_existing_connection_value_when_migrating() -> None:
    config = ConfigObj({'main': {}, 'connection': {'default_ssl_mode': 'off'}})
    config_wo = ConfigObj({'main': {'ssl_mode': 'on'}, 'connection': {'default_ssl_mode': 'off'}})

    ssl_mode, warning = normalize_ssl_mode(config, config_wo)

    assert ssl_mode == 'off'
    assert warning == (
        'Mycli 2.0 migration: automatically moving ssl_mode under [main] to default_ssl_mode under [connection] in ~/.myclirc .'
        '\nBut connection.default_ssl_mode already existed, with the value: "off".'
    )
    assert config_wo['connection']['default_ssl_mode'] == 'off'
    assert 'ssl_mode' not in config_wo['main']


def test_normalize_ssl_mode_reports_invalid_values() -> None:
    config = ConfigObj({'main': {'ssl_mode': 'required'}, 'connection': {'default_ssl_mode': 'required'}})
    config_wo = ConfigObj()

    ssl_mode, warning = normalize_ssl_mode(config, config_wo)

    assert ssl_mode is None
    assert warning == 'Invalid config option provided for ssl_mode (required); ignoring.'


def test_destructive_keywords_from_config_splits_non_empty_words() -> None:
    config = ConfigObj({'main': {'destructive_keywords': 'DROP  DELETE  UPDATE'}})

    assert destructive_keywords_from_config(config) == ['DROP', 'DELETE', 'UPDATE']


def test_destructive_keywords_from_config_uses_default() -> None:
    config = ConfigObj({'main': {}})

    assert destructive_keywords_from_config(config) == ['DROP', 'SHUTDOWN', 'DELETE', 'TRUNCATE', 'ALTER', 'UPDATE']


@pytest.mark.parametrize(
    ('llm_config', 'expected'),
    [
        ({'prompt_field_truncate': '12', 'prompt_section_truncate': '34'}, (12, 34)),
        ({'prompt_field_truncate': 'abc', 'prompt_section_truncate': '-1'}, (0, 0)),
        ({}, (0, 0)),
    ],
)
def test_llm_prompt_truncation_reads_positive_integer_strings(
    llm_config: dict[str, str],
    expected: tuple[int, int],
) -> None:
    config = ConfigObj({'main': {}, 'llm': llm_config})

    assert llm_prompt_truncation(config) == expected


def test_llm_prompt_truncation_handles_missing_llm_section() -> None:
    assert llm_prompt_truncation(ConfigObj({'main': {}})) == (0, 0)


def test_read_mylogin_cnf_reads_login_path_only() -> None:
    app_state = AppState(login_path='work')
    cnf = ConfigObj({
        'client': {'user': 'client-user'},
        'work': {'password': 'work-pass'},
        'clienttest': {'host': 'client-test-host'},
        'worktest': {'socket': 'work-test-socket'},
    })

    configuration = app_state.read_mylogin_cnf(cnf)

    assert configuration == {
        'user': None,
        'password': 'work-pass',
        'host': None,
        'port': None,
        'socket': None,
    }


def test_read_mylogin_cnf_strips_quotes() -> None:
    app_state = AppState(login_path='work')
    cnf = ConfigObj({
        'work': {'password': '"work-pass"'},
    })

    configuration = app_state.read_mylogin_cnf(cnf)

    assert configuration == {
        'user': None,
        'password': 'work-pass',
        'host': None,
        'port': None,
        'socket': None,
    }



from __future__ import annotations

from types import SimpleNamespace
from typing import Any

import pytest

from mycli import cli_runner, main
from mycli.password_sources import (
    KNOWN_PASSWORD_SOURCES,
)


class DummyLogger:
    def __init__(self) -> None:
        self.debug_calls: list[tuple[tuple[Any, ...], dict[str, Any]]] = []

    def debug(self, *args: Any, **kwargs: Any) -> None:
        self.debug_calls.append((args, kwargs))


class DummyMyCli:
    def __init__(
        self,
        *,
        config: dict[str, Any] | None = None,
        my_cnf: dict[str, Any] | None = None,
        config_without_package_defaults: dict[str, Any] | None = None,
    ) -> None:
        self.config = config or default_config()
        self.my_cnf = my_cnf or {'client': {}, 'mysqld': {}}
        self.config_without_package_defaults = config_without_package_defaults or {}
        self.default_keepalive_ticks = 5
        self.ssl_mode: str | None = None
        self.logger = DummyLogger()
        self.dsn_alias: str | None = None
        self.ssh_tunnel: Any = None
        self.connect_calls: list[dict[str, Any]] = []
        self.run_cli_called = False
        self.close_called = False

    def connect(self, **kwargs: Any) -> None:
        self.connect_calls.append(dict(kwargs))

    def run_cli(self) -> None:
        self.run_cli_called = True

    def close(self) -> None:
        if getattr(self, 'ssh_tunnel', None) is not None:
            self.ssh_tunnel.close()
        self.close_called = True


def default_config() -> dict[str, Any]:
    return {
        'main': {'use_keyring': 'false'},
        'connection': {'default_keepalive_ticks': 0},
        'vault_beta': {},
        'alias_dsn': {},
        'init-commands': {},
        'alias_dsn.init-commands': {},
    }


def make_cli_args() -> main.CliArgs:
    cli_args = main.CliArgs()
    cli_args.format = None
    return cli_args


def run_with_client(
    monkeypatch: pytest.MonkeyPatch,
    cli_args: main.CliArgs,
    client: DummyMyCli,
) -> DummyMyCli:
    monkeypatch.setattr(main, 'preprocess_cli_args', lambda args, scheme_validator: 4)
    monkeypatch.setattr(cli_runner.sys, 'stdin', SimpleNamespace(isatty=lambda: True))
    monkeypatch.setattr(cli_runner.sys.stderr, 'isatty', lambda: False)
    cli_runner.run_from_cli_args(cli_args, lambda **_kwargs: client)
    return client


def resolve_connect_password(connect_call: dict[str, Any]) -> tuple[str, str | int] | None:
    selected = connect_call['password_candidates'].resolve(KNOWN_PASSWORD_SOURCES)
    if selected is None:
        return None
    return selected.source, selected.value


def test_expand_dsn_alias_env_var_returns_none() -> None:
    assert cli_runner.expand_dsn_alias_env_var(None, 'prod') is None


def test_split_dsn_netloc_handles_user_without_password() -> None:
    assert cli_runner.split_dsn_netloc('user@host:3306') == ('user', None, 'host', '3306')


def test_split_dsn_netloc_handles_empty_host() -> None:
    assert cli_runner.split_dsn_netloc('user:pass@') == ('user', 'pass', None, None)


def test_split_dsn_netloc_handles_bracketed_ipv6_host() -> None:
    assert cli_runner.split_dsn_netloc('user:pass@[::1]:3306') == ('user', 'pass', '::1', '3306')


def test_expand_dsn_alias_env_vars_rejects_non_integer_port(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv('MYCLI_TEST_DSN_PORT', 'not-an-int')

    with pytest.raises(cli_runner.DsnAliasEnvVarError) as excinfo:
        cli_runner.expand_dsn_alias_env_vars('mysql://user:pass@host:${MYCLI_TEST_DSN_PORT}/db', 'prod')

    assert str(excinfo.value) == 'Port in DSN alias prod must be an integer.'


def test_run_from_cli_args_checkup_exits_zero(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.checkup = True
    client = DummyMyCli()
    checkup_calls: list[DummyMyCli] = []
    monkeypatch.setattr(cli_runner, 'main_checkup', lambda value: checkup_calls.append(value))
    monkeypatch.setattr(main, 'preprocess_cli_args', lambda args, scheme_validator: 0)

    with pytest.raises(SystemExit) as excinfo:
        cli_runner.run_from_cli_args(cli_args, lambda **_kwargs: client)

    assert excinfo.value.code == 0
    assert checkup_calls == [client]


@pytest.mark.parametrize(
    ('csv', 'table', 'format_name', 'message'),
    (
        (True, False, 'table', 'Conflicting --csv and --format arguments.'),
        (False, True, 'csv', 'Conflicting --table and --format arguments.'),
    ),
)
def test_run_from_cli_args_rejects_conflicting_format_flags(
    monkeypatch: pytest.MonkeyPatch,
    csv: bool,
    table: bool,
    format_name: str,
    message: str,
) -> None:
    cli_args = make_cli_args()
    cli_args.csv = csv
    cli_args.table = table
    cli_args.format = format_name
    secho_calls: list[tuple[str, dict[str, Any]]] = []
    monkeypatch.setattr(cli_runner.click, 'secho', lambda text, **kwargs: secho_calls.append((text, kwargs)))
    monkeypatch.setattr(main, 'preprocess_cli_args', lambda args, scheme_validator: 0)

    with pytest.raises(SystemExit) as excinfo:
        cli_runner.run_from_cli_args(cli_args, lambda **_kwargs: DummyMyCli())

    assert excinfo.value.code == 1
    assert secho_calls == [(message, {'err': True, 'fg': 'red'})]


def test_run_from_cli_args_treats_database_as_dsn_alias(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.database = 'prod'
    client = DummyMyCli(
        config={
            **default_config(),
            'alias_dsn': {'prod': 'mysql://u:p@h/db'},
        }
    )

    run_with_client(monkeypatch, cli_args, client)

    assert client.dsn_alias == 'prod'
    connect_call = client.connect_calls[-1]
    assert connect_call['user'] == 'u'
    assert resolve_connect_password(connect_call) == ('dsn', 'p')
    assert connect_call['host'] == 'h'
    assert connect_call['database'] == 'db'


def test_run_from_cli_args_password_file_prevents_positional_dsn_alias(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.database = 'prod'
    cli_args.password_file = 'secret.txt'
    client = DummyMyCli(
        config={
            **default_config(),
            'alias_dsn': {'prod': 'mysql://u:p@h/db'},
        }
    )
    password_file_calls: list[str | None] = []

    def read_password_file(password_file: str | None) -> str:
        password_file_calls.append(password_file)
        return 'file-secret'

    monkeypatch.setattr(main, 'get_password_from_file', read_password_file)

    run_with_client(monkeypatch, cli_args, client)

    connect_call = client.connect_calls[-1]
    assert client.dsn_alias is None
    assert connect_call['database'] == 'prod'
    assert resolve_connect_password(connect_call) == ('file', 'file-secret')
    assert password_file_calls == ['secret.txt']


def test_run_from_cli_args_mysql_pwd_prevents_positional_dsn_alias(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.database = 'prod'
    client = DummyMyCli(
        config={
            **default_config(),
            'alias_dsn': {'prod': 'mysql://u:p@h/db'},
        }
    )
    monkeypatch.setenv('MYSQL_PWD', 'environment-secret')

    run_with_client(monkeypatch, cli_args, client)

    connect_call = client.connect_calls[-1]
    assert client.dsn_alias is None
    assert connect_call['database'] == 'prod'
    assert resolve_connect_password(connect_call) == ('environment', 'environment-secret')


def test_run_from_cli_args_keeps_empty_cli_password_over_dsn(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'mysql://user:dsn-secret@host/db'
    cli_args.password = ''
    client = DummyMyCli()

    run_with_client(monkeypatch, cli_args, client)

    assert resolve_connect_password(client.connect_calls[-1]) == ('literal', '')


def test_run_from_cli_args_preserves_cli_password_prompt(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.password = cli_runner.EMPTY_PASSWORD_FLAG_SENTINEL
    client = DummyMyCli()

    run_with_client(monkeypatch, cli_args, client)

    assert resolve_connect_password(client.connect_calls[-1]) == (
        'prompt',
        cli_runner.EMPTY_PASSWORD_FLAG_SENTINEL,
    )


def test_run_from_cli_args_leaves_dsn_alias_env_vars_disabled_by_default(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'prod'
    monkeypatch.setenv('MYCLI_TEST_DSN_USER', 'env_user')
    client = DummyMyCli(
        config={
            **default_config(),
            'alias_dsn': {'prod': 'mysql://${MYCLI_TEST_DSN_USER}:pass@host:3306/db'},
        }
    )

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['user'] == '${MYCLI_TEST_DSN_USER}'


def test_run_from_cli_args_expands_whole_dsn_alias_env_vars_when_enabled(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'prod'
    monkeypatch.setenv('MYCLI_TEST_DSN_USER', 'env_user')
    monkeypatch.setenv('MYCLI_TEST_DSN_PASSWORD', 'env_pass')
    monkeypatch.setenv('MYCLI_TEST_DSN_HOST', 'env-host')
    monkeypatch.setenv('MYCLI_TEST_DSN_PORT', '3308')
    monkeypatch.setenv('MYCLI_TEST_DSN_DATABASE', 'env_db')
    monkeypatch.setenv('MYCLI_TEST_DSN_CHARSET', 'utf8mb4')
    monkeypatch.setenv('MYCLI_TEST_DSN_KEEPALIVE', '9')
    config = default_config()
    config['main'] = {**config['main'], 'expand_dsn_alias_env_vars': 'true'}
    config['alias_dsn'] = {
        'prod': (
            'mysql://${MYCLI_TEST_DSN_USER}:${MYCLI_TEST_DSN_PASSWORD}'
            '@${MYCLI_TEST_DSN_HOST}:${MYCLI_TEST_DSN_PORT}/${MYCLI_TEST_DSN_DATABASE}'
            '?character_set=${MYCLI_TEST_DSN_CHARSET}&keepalive_ticks=${MYCLI_TEST_DSN_KEEPALIVE}'
        )
    }
    client = DummyMyCli(config=config)

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['user'] == 'env_user'
    assert resolve_connect_password(client.connect_calls[-1]) == ('dsn', 'env_pass')
    assert client.connect_calls[-1]['host'] == 'env-host'
    assert client.connect_calls[-1]['port'] == 3308
    assert client.connect_calls[-1]['database'] == 'env_db'
    assert client.connect_calls[-1]['character_set'] == 'utf8mb4'
    assert client.connect_calls[-1]['keepalive_ticks'] == 9


def test_run_from_cli_args_expands_dsn_alias_ssh_jump_env_var_when_enabled(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'prod'
    monkeypatch.setenv('MYCLI_TEST_DSN_SSH_JUMP', 'env-bastion')
    config = default_config()
    config['main'] = {**config['main'], 'expand_dsn_alias_env_vars': 'true'}
    config['alias_dsn'] = {'prod': 'mysql://user@host/db?ssh_jump=${MYCLI_TEST_DSN_SSH_JUMP}'}
    client = DummyMyCli(config=config)

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['ssh_jump'] == 'env-bastion'


def test_run_from_cli_args_does_not_expand_partial_values_or_query_keys(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'prod'
    monkeypatch.setenv('MYCLI_TEST_DSN_USER', 'env_user')
    monkeypatch.setenv('MYCLI_TEST_DSN_QUERY_KEY', 'character_set')
    config = default_config()
    config['main'] = {**config['main'], 'expand_dsn_alias_env_vars': 'true'}
    config['alias_dsn'] = {
        'prod': ('mysql://user-${MYCLI_TEST_DSN_USER}:pass@host:3306/db?${MYCLI_TEST_DSN_QUERY_KEY}=utf8mb4&character_set=utf8')
    }
    client = DummyMyCli(config=config)

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['user'] == 'user-${MYCLI_TEST_DSN_USER}'
    assert client.connect_calls[-1]['character_set'] == 'utf8'


def test_run_from_cli_args_does_not_expand_unbraced_dsn_alias_env_vars(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'prod'
    monkeypatch.setenv('MYCLI_TEST_DSN_USER', 'env_user')
    config = default_config()
    config['main'] = {**config['main'], 'expand_dsn_alias_env_vars': 'true'}
    config['alias_dsn'] = {'prod': 'mysql://$MYCLI_TEST_DSN_USER:pass@host:3306/db'}
    client = DummyMyCli(config=config)

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['user'] == '$MYCLI_TEST_DSN_USER'


def test_run_from_cli_args_warns_about_unknown_dsn_query_parameter(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'mysql://user@host/db?unknown=value'
    client = DummyMyCli()
    secho_calls: list[tuple[str, dict[str, Any]]] = []
    monkeypatch.setattr(cli_runner.click, 'secho', lambda text, **kwargs: secho_calls.append((text, kwargs)))

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['database'] == 'db'
    assert secho_calls == [
        (
            'Warning: Ignored unknown DSN URI query parameters: unknown.',
            {'err': True, 'fg': 'yellow'},
        )
    ]


def test_run_from_cli_args_warns_once_about_sorted_unknown_dsn_query_parameters(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'mysql://user@host/db?zzz=1&aaa&character_set=utf8&aaa=2'
    client = DummyMyCli()
    secho_calls: list[tuple[str, dict[str, Any]]] = []
    monkeypatch.setattr(cli_runner.click, 'secho', lambda text, **kwargs: secho_calls.append((text, kwargs)))

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['character_set'] == 'utf8'
    assert secho_calls == [
        (
            'Warning: Ignored unknown DSN URI query parameters: aaa, zzz.',
            {'err': True, 'fg': 'yellow'},
        )
    ]


def test_run_from_cli_args_warns_about_unknown_alias_dsn_query_parameter_with_env_expansion(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'prod'
    monkeypatch.setenv('MYCLI_TEST_DSN_CHARSET', 'utf8')
    config = default_config()
    config['main'] = {**config['main'], 'expand_dsn_alias_env_vars': 'true'}
    config['alias_dsn'] = {'prod': 'mysql://user@host/db?unknown=value&character_set=${MYCLI_TEST_DSN_CHARSET}'}
    client = DummyMyCli(config=config)
    secho_calls: list[tuple[str, dict[str, Any]]] = []
    monkeypatch.setattr(cli_runner.click, 'secho', lambda text, **kwargs: secho_calls.append((text, kwargs)))

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['character_set'] == 'utf8'
    assert secho_calls == [
        (
            'Warning: Ignored unknown DSN URI query parameters: unknown.',
            {'err': True, 'fg': 'yellow'},
        )
    ]


def test_run_from_cli_args_reports_missing_dsn_alias_env_var(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'prod'
    config = default_config()
    config['main'] = {**config['main'], 'expand_dsn_alias_env_vars': 'true'}
    config['alias_dsn'] = {'prod': 'mysql://${MYCLI_TEST_MISSING_DSN_USER}:pass@host:3306/db'}
    client = DummyMyCli(config=config)
    secho_calls: list[tuple[str, dict[str, Any]]] = []
    monkeypatch.setattr(cli_runner.click, 'secho', lambda text, **kwargs: secho_calls.append((text, kwargs)))

    with pytest.raises(SystemExit) as excinfo:
        run_with_client(monkeypatch, cli_args, client)

    assert excinfo.value.code == 1
    assert secho_calls == [
        (
            'Environment variable MYCLI_TEST_MISSING_DSN_USER referenced by DSN alias prod is not set.',
            {'err': True, 'fg': 'red'},
        )
    ]
    assert client.connect_calls == []


def test_run_from_cli_args_reports_missing_dsn(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'missing'
    secho_calls: list[tuple[str, dict[str, Any]]] = []
    monkeypatch.setattr(cli_runner, 'is_valid_connection_scheme', lambda value: (False, None))
    monkeypatch.setattr(cli_runner.click, 'secho', lambda text, **kwargs: secho_calls.append((text, kwargs)))
    monkeypatch.setattr(main, 'preprocess_cli_args', lambda args, scheme_validator: 0)

    with pytest.raises(SystemExit) as excinfo:
        cli_runner.run_from_cli_args(cli_args, lambda **_kwargs: DummyMyCli())

    assert excinfo.value.code == 1
    assert secho_calls == [
        (
            'Could not find the specified DSN in the config file. Please check the "[alias_dsn]" section in your myclirc.',
            {'err': True, 'fg': 'red'},
        )
    ]


def test_run_from_cli_args_rejects_unknown_positional_dsn_scheme(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.database = 'ssh://user@example.com/db'
    secho_calls: list[tuple[str, dict[str, Any]]] = []
    monkeypatch.setattr(cli_runner.click, 'secho', lambda text, **kwargs: secho_calls.append((text, kwargs)))

    with pytest.raises(SystemExit) as excinfo:
        run_with_client(monkeypatch, cli_args, DummyMyCli())

    assert excinfo.value.code == 1
    assert secho_calls == [
        (
            'Error: Unknown connection scheme provided for DSN URI (ssh://)',
            {'err': True, 'fg': 'red'},
        )
    ]


def test_run_from_cli_args_rejects_unknown_alias_dsn_scheme(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'legacy_ssh'
    client = DummyMyCli(
        config={
            **default_config(),
            'alias_dsn': {'legacy_ssh': 'ssh://user@example.com/db'},
        }
    )
    secho_calls: list[tuple[str, dict[str, Any]]] = []
    monkeypatch.setattr(cli_runner.click, 'secho', lambda text, **kwargs: secho_calls.append((text, kwargs)))

    with pytest.raises(SystemExit) as excinfo:
        run_with_client(monkeypatch, cli_args, client)

    assert excinfo.value.code == 1
    assert secho_calls == [
        (
            'Error: Unknown connection scheme provided for DSN URI (ssh://)',
            {'err': True, 'fg': 'red'},
        )
    ]


def test_run_from_cli_args_accepts_mysql_plus_dsn_scheme(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'mysql+pymysql://user:pass@host:3306/db'
    client = DummyMyCli()

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['user'] == 'user'
    assert resolve_connect_password(client.connect_calls[-1]) == ('dsn', 'pass')
    assert client.connect_calls[-1]['host'] == 'host'
    assert client.connect_calls[-1]['port'] == 3306
    assert client.connect_calls[-1]['database'] == 'db'


def test_run_from_cli_args_maps_dsn_ssh_jump_parameter(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'mysql://user@host/db?ssh_jump=bastion'
    client = DummyMyCli()

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['ssh_jump'] == 'bastion'


def test_run_from_cli_args_does_not_warn_about_known_dsn_query_parameter(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'mysql://user@host/db?ssh_jump=bastion'
    client = DummyMyCli()
    secho_calls: list[tuple[str, dict[str, Any]]] = []
    monkeypatch.setattr(cli_runner.click, 'secho', lambda text, **kwargs: secho_calls.append((text, kwargs)))

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['ssh_jump'] == 'bastion'
    assert secho_calls == []


def test_run_from_cli_args_prefers_cli_ssh_jump_over_dsn_parameter(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'mysql://user@host/db?ssh_jump=dsn-bastion'
    cli_args.ssh_jump = 'cli-bastion'
    client = DummyMyCli()

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['ssh_jump'] == 'cli-bastion'


def test_run_from_cli_args_maps_dsn_ssl_parameters(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = (
        'mysql://user:pass@host:3306/db?ssl_mode=on&ssl_ca=~/ca.pem&ssl_capath=/capath'
        '&ssl_cert=~/cert.pem&ssl_key=~/key.pem&ssl_cipher=AES256&tls_version=TLSv1.3'
        '&ssl_verify_server_cert=true'
    )
    client = DummyMyCli()
    secho_calls: list[str] = []
    monkeypatch.setattr(cli_runner.click, 'secho', lambda text, **_kwargs: secho_calls.append(text))

    run_with_client(monkeypatch, cli_args, client)

    ssl = client.connect_calls[-1]['ssl']
    assert ssl == {
        'mode': 'on',
        'ca': cli_runner.os.path.expanduser('~/ca.pem'),
        'capath': '/capath',
        'cert': cli_runner.os.path.expanduser('~/cert.pem'),
        'key': cli_runner.os.path.expanduser('~/key.pem'),
        'cipher': 'AES256',
        'tls_version': 'TLSv1.3',
        'check_hostname': True,
    }


def test_run_from_cli_args_merges_global_list_and_alias_scalar_init_commands(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'prod'
    cli_args.init_command = 'set cli=1'
    client = DummyMyCli(
        config={
            **default_config(),
            'alias_dsn': {'prod': 'mysql://u:p@h/db'},
            'init-commands': {'first': ['set global=1', 'set global=2']},
            'alias_dsn.init-commands': {'prod': 'set alias=1'},
        }
    )

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['init_command'] == 'set global=1; set global=2; set alias=1; set cli=1'


def test_run_from_cli_args_resets_keyring(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.use_keyring = 'reset'
    client = DummyMyCli()

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['use_keyring'] is True
    assert client.connect_calls[-1]['reset_keyring'] is True


def test_run_from_cli_args_uses_explicit_keyring_flag(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.use_keyring = 'true'
    client = DummyMyCli()

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['use_keyring'] is True
    assert client.connect_calls[-1]['reset_keyring'] is False


def test_run_from_cli_args_reads_password_from_vault_when_password_is_missing(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.user = 'existing-user'
    cli_args.vault_secret = 'database/prod'
    client = DummyMyCli(
        config={
            **default_config(),
            'vault_beta': {
                'vault_executable': '/opt/bin/vault',
                'address': 'https://vault.config',
                'default_mount': 'kv',
                'default_password_field': 'mysql_password',
            },
        }
    )
    vault_calls: list[dict[str, str | None]] = []

    def fake_get_field_from_vault(field: str, secret: str, **kwargs: str | None) -> str:
        vault_calls.append({'field': field, 'secret': secret, **kwargs})
        return 'vault-secret'

    monkeypatch.delenv('VAULT_ADDR', raising=False)
    monkeypatch.setattr(cli_runner, 'get_field_from_vault', fake_get_field_from_vault)

    run_with_client(monkeypatch, cli_args, client)

    assert resolve_connect_password(client.connect_calls[-1]) == ('vault', 'vault-secret')
    assert vault_calls == [
        {
            'secret': 'database/prod',
            'executable': '/opt/bin/vault',
            'field': 'mysql_password',
            'mount': 'kv',
            'address': 'https://vault.config',
        }
    ]


def test_run_from_cli_args_reads_username_from_vault_when_user_is_missing(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.password = 'existing-password'
    cli_args.vault_secret = 'database/prod'
    cli_args.vault_username_field = 'mysql_username'
    client = DummyMyCli(
        config={
            **default_config(),
            'vault_beta': {
                'vault_executable': '/opt/bin/vault',
                'address': 'https://vault.config',
                'default_mount': 'kv',
            },
        }
    )
    vault_calls: list[dict[str, str | None]] = []

    def fake_get_field_from_vault(field: str, secret: str, **kwargs: str | None) -> str:
        vault_calls.append({'field': field, 'secret': secret, **kwargs})
        return 'vault-user'

    monkeypatch.delenv('VAULT_ADDR', raising=False)
    monkeypatch.setattr(cli_runner, 'get_field_from_vault', fake_get_field_from_vault)

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['user'] == 'vault-user'
    assert client.connect_calls[-1]['vault_username_from_vault'] is True
    assert vault_calls == [
        {
            'secret': 'database/prod',
            'executable': '/opt/bin/vault',
            'field': 'mysql_username',
            'mount': 'kv',
            'address': 'https://vault.config',
        }
    ]


def test_run_from_cli_args_prefers_dsn_user_over_vault_username(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'mysql://dsn-user@host/db'
    cli_args.password = 'existing-password'
    cli_args.vault_secret = 'database/prod'
    cli_args.vault_username_field = 'mysql_username'
    client = DummyMyCli()
    vault_calls: list[dict[str, str | None]] = []

    def fake_get_field_from_vault(field: str, secret: str, **kwargs: str | None) -> str:
        vault_calls.append({'field': field, 'secret': secret, **kwargs})
        return 'vault-user'

    monkeypatch.setattr(cli_runner, 'get_field_from_vault', fake_get_field_from_vault)

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['user'] == 'dsn-user'
    assert client.connect_calls[-1]['vault_username_from_vault'] is False
    assert vault_calls == []


def test_run_from_cli_args_reports_vault_username_error(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.password = 'existing-password'
    cli_args.vault_secret = 'database/prod'
    cli_args.vault_username_field = 'mysql_username'
    client = DummyMyCli()
    secho_calls: list[tuple[str, dict[str, Any]]] = []
    monkeypatch.setattr(cli_runner.click, 'secho', lambda text, **kwargs: secho_calls.append((text, kwargs)))
    monkeypatch.setattr(
        cli_runner,
        'get_field_from_vault',
        lambda *_args, **_kwargs: (_ for _ in ()).throw(cli_runner.VaultError('permission denied')),
    )

    with pytest.raises(SystemExit) as excinfo:
        run_with_client(monkeypatch, cli_args, client)

    assert excinfo.value.code == 1
    assert client.connect_calls == []
    assert secho_calls == [('Error reading username from Vault: permission denied', {'err': True, 'fg': 'red'})]


def test_run_from_cli_args_prefers_dsn_password_over_vault(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'mysql://user:dsn-secret@host/db'
    cli_args.vault_secret = 'database/prod'
    client = DummyMyCli()
    vault_calls: list[dict[str, str | None]] = []

    def fake_get_field_from_vault(field: str, secret: str, **kwargs: str | None) -> str:
        vault_calls.append({'field': field, 'secret': secret, **kwargs})
        return 'vault-secret'

    monkeypatch.setattr(cli_runner, 'get_field_from_vault', fake_get_field_from_vault)

    run_with_client(monkeypatch, cli_args, client)

    assert resolve_connect_password(client.connect_calls[-1]) == ('dsn', 'dsn-secret')
    assert vault_calls == []


def test_run_from_cli_args_prefers_vault_cli_values_and_env_address(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.user = 'existing-user'
    cli_args.vault_secret = 'database/prod'
    cli_args.vault_mount = 'cli-mount'
    cli_args.vault_password_field = 'cli-field'
    client = DummyMyCli(
        config={
            **default_config(),
            'vault_beta': {
                'vault_executable': '/opt/bin/vault',
                'address': 'https://vault.config',
                'default_mount': 'config-mount',
                'default_password_field': 'config-field',
            },
        }
    )
    vault_calls: list[dict[str, str | None]] = []
    monkeypatch.setenv('VAULT_ADDR', 'https://vault.env')

    def fake_get_field_from_vault(field: str, secret: str, **kwargs: str | None) -> str:
        vault_calls.append({'field': field, 'secret': secret, **kwargs})
        return 'vault-secret'

    monkeypatch.setattr(cli_runner, 'get_field_from_vault', fake_get_field_from_vault)

    run_with_client(monkeypatch, cli_args, client)

    assert resolve_connect_password(client.connect_calls[-1]) == ('vault', 'vault-secret')
    assert client.connect_calls[-1]['vault_username_from_vault'] is False
    assert vault_calls == [
        {
            'secret': 'database/prod',
            'executable': '/opt/bin/vault',
            'field': 'cli-field',
            'mount': 'cli-mount',
            'address': 'https://vault.env',
        }
    ]


def test_run_from_cli_args_prefers_vault_cli_address_over_env(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.user = 'existing-user'
    cli_args.vault_secret = 'database/prod'
    cli_args.vault_address = 'https://vault.cli'
    client = DummyMyCli()
    vault_calls: list[dict[str, str | None]] = []
    monkeypatch.setenv('VAULT_ADDR', 'https://vault.env')

    def fake_get_field_from_vault(field: str, secret: str, **kwargs: str | None) -> str:
        vault_calls.append({'field': field, 'secret': secret, **kwargs})
        return 'vault-secret'

    monkeypatch.setattr(cli_runner, 'get_field_from_vault', fake_get_field_from_vault)

    run_with_client(monkeypatch, cli_args, client)

    assert resolve_connect_password(client.connect_calls[-1]) == ('vault', 'vault-secret')
    assert vault_calls[-1]['address'] == 'https://vault.cli'


def test_run_from_cli_args_reports_vault_error(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.user = 'existing-user'
    cli_args.vault_secret = 'database/prod'
    client = DummyMyCli()
    secho_calls: list[tuple[str, dict[str, Any]]] = []
    monkeypatch.setattr(cli_runner.click, 'secho', lambda text, **kwargs: secho_calls.append((text, kwargs)))
    monkeypatch.setattr(
        cli_runner,
        'get_field_from_vault',
        lambda *_args, **_kwargs: (_ for _ in ()).throw(cli_runner.VaultError('permission denied')),
    )

    run_with_client(monkeypatch, cli_args, client)

    with pytest.raises(SystemExit) as excinfo:
        resolve_connect_password(client.connect_calls[-1])

    assert excinfo.value.code == 1
    assert secho_calls == [('Error reading password from Vault: permission denied', {'err': True, 'fg': 'red'})]


def test_run_from_cli_args_passes_ssh_options_to_connect(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.ssh_options = '-o Compression=yes'
    client = DummyMyCli()

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['ssh_cli_options'] == '-o Compression=yes'


@pytest.mark.parametrize(
    ('ssh_connection', 'expected_use_keyring'),
    (
        (None, True),
        ('client-ip client-port server-ip server-port', False),
    ),
)
def test_run_from_cli_args_uses_auto_keyring_flag(
    monkeypatch: pytest.MonkeyPatch,
    ssh_connection: str | None,
    expected_use_keyring: bool,
) -> None:
    cli_args = make_cli_args()
    cli_args.use_keyring = 'auto'
    client = DummyMyCli()
    if ssh_connection is None:
        monkeypatch.delenv('SSH_CONNECTION', raising=False)
    else:
        monkeypatch.setenv('SSH_CONNECTION', ssh_connection)

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['use_keyring'] is expected_use_keyring
    assert client.connect_calls[-1]['reset_keyring'] is False


@pytest.mark.parametrize(
    ('ssh_connection', 'expected_use_keyring'),
    (
        (None, True),
        ('client-ip client-port server-ip server-port', False),
    ),
)
def test_run_from_cli_args_uses_auto_keyring_config(
    monkeypatch: pytest.MonkeyPatch,
    ssh_connection: str | None,
    expected_use_keyring: bool,
) -> None:
    cli_args = make_cli_args()
    config = default_config()
    config['main'] = {**config['main'], 'use_keyring': 'auto'}
    client = DummyMyCli(config=config)
    if ssh_connection is None:
        monkeypatch.delenv('SSH_CONNECTION', raising=False)
    else:
        monkeypatch.setenv('SSH_CONNECTION', ssh_connection)

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['use_keyring'] is expected_use_keyring
    assert client.connect_calls[-1]['reset_keyring'] is False


@pytest.mark.parametrize(
    ('flag_name', 'expected_format'),
    (
        ('csv', 'csv'),
        ('table', 'table'),
    ),
)
def test_run_from_cli_args_sets_legacy_format_flags(
    monkeypatch: pytest.MonkeyPatch,
    flag_name: str,
    expected_format: str,
) -> None:
    cli_args = make_cli_args()
    setattr(cli_args, flag_name, True)
    client = DummyMyCli()

    run_with_client(monkeypatch, cli_args, client)

    assert cli_args.format == expected_format


def test_run_from_cli_args_list_dsn_exits_with_mode_result(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.list_dsn = True
    client = DummyMyCli()
    list_dsn_calls: list[DummyMyCli] = []
    monkeypatch.setattr(main, 'preprocess_cli_args', lambda args, scheme_validator: 0)

    def fake_main_list_dsn(value: DummyMyCli) -> int:
        list_dsn_calls.append(value)
        return 7

    monkeypatch.setattr(cli_runner, 'main_list_dsn', fake_main_list_dsn)

    with pytest.raises(SystemExit) as excinfo:
        cli_runner.run_from_cli_args(cli_args, lambda **_kwargs: client)

    assert excinfo.value.code == 7
    assert list_dsn_calls == [client]


def test_run_from_cli_args_maps_dsn_socket_parameter(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'mysql://user@host/db?socket=/tmp/mysql.sock'
    client = DummyMyCli()

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['socket'] == '/tmp/mysql.sock'


def test_run_from_cli_args_maps_dsn_vault_parameters(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.user = 'existing-user'
    cli_args.password = 'existing-password'
    cli_args.dsn = (
        'mysql://host/db?vault_address=https%3A%2F%2Fvault.example.com&vault_mount=kv'
        '&vault_secret=database%2Fprod&vault_password_field=mysql_password'
        '&vault_username_field=mysql_username'
    )
    client = DummyMyCli()

    run_with_client(monkeypatch, cli_args, client)

    connect_call = client.connect_calls[-1]
    assert connect_call['vault_address'] == 'https://vault.example.com'
    assert connect_call['vault_mount'] == 'kv'
    assert connect_call['vault_secret'] == 'database/prod'
    assert connect_call['vault_password_field'] == 'mysql_password'
    assert connect_call['vault_username_field'] == 'mysql_username'


def test_run_from_cli_args_uses_no_ssl_for_auto_ssl_over_socket(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.socket = '/tmp/mysql.sock'
    client = DummyMyCli()
    client.ssl_mode = 'auto'

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['ssl'] is None


def test_run_from_cli_args_merges_scalar_global_and_alias_list_init_commands(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.dsn = 'prod'
    client = DummyMyCli(
        config={
            **default_config(),
            'alias_dsn': {'prod': 'mysql://u:p@h/db'},
            'init-commands': {'global': 'set global=1'},
            'alias_dsn.init-commands': {'prod': ['set alias=1', 'set alias=2']},
        }
    )

    run_with_client(monkeypatch, cli_args, client)

    assert client.connect_calls[-1]['init_command'] == 'set global=1; set alias=1; set alias=2'


def test_run_from_cli_args_execute_mode_exits_with_mode_result(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    cli_args.execute = 'select 1'
    client = DummyMyCli()
    execute_calls: list[tuple[DummyMyCli, main.CliArgs]] = []
    monkeypatch.setattr(main, 'preprocess_cli_args', lambda args, scheme_validator: 0)
    monkeypatch.setattr(cli_runner.sys, 'stdin', SimpleNamespace(isatty=lambda: True))
    monkeypatch.setattr(cli_runner.sys.stderr, 'isatty', lambda: False)

    def fake_main_execute_from_cli(mycli: DummyMyCli, args: main.CliArgs) -> int:
        execute_calls.append((mycli, args))
        return 11

    monkeypatch.setattr(cli_runner, 'main_execute_from_cli', fake_main_execute_from_cli)

    with pytest.raises(SystemExit) as excinfo:
        cli_runner.run_from_cli_args(cli_args, lambda **_kwargs: client)

    assert excinfo.value.code == 11
    assert execute_calls == [(client, cli_args)]
    assert client.close_called is True


def test_run_from_cli_args_batch_with_progress_exits_with_mode_result(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.batch = 'input.sql'
    cli_args.progress = True
    client = DummyMyCli()
    batch_calls: list[tuple[DummyMyCli, main.CliArgs]] = []
    monkeypatch.setattr(main, 'preprocess_cli_args', lambda args, scheme_validator: 0)
    monkeypatch.setattr(cli_runner.sys, 'stdin', SimpleNamespace(isatty=lambda: True))
    monkeypatch.setattr(cli_runner.sys.stderr, 'isatty', lambda: True)

    def fake_main_batch_with_progress_bar(mycli: DummyMyCli, args: main.CliArgs) -> int:
        batch_calls.append((mycli, args))
        return 12

    monkeypatch.setattr(cli_runner, 'main_batch_with_progress_bar', fake_main_batch_with_progress_bar)

    with pytest.raises(SystemExit) as excinfo:
        cli_runner.run_from_cli_args(cli_args, lambda **_kwargs: client)

    assert excinfo.value.code == 12
    assert batch_calls == [(client, cli_args)]
    assert client.close_called is True


def test_run_from_cli_args_batch_without_progress_exits_with_mode_result(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cli_args = make_cli_args()
    cli_args.batch = 'input.sql'
    client = DummyMyCli()
    batch_calls: list[tuple[DummyMyCli, main.CliArgs]] = []
    monkeypatch.setattr(main, 'preprocess_cli_args', lambda args, scheme_validator: 0)
    monkeypatch.setattr(cli_runner.sys, 'stdin', SimpleNamespace(isatty=lambda: True))
    monkeypatch.setattr(cli_runner.sys.stderr, 'isatty', lambda: False)

    def fake_main_batch_without_progress_bar(mycli: DummyMyCli, args: main.CliArgs) -> int:
        batch_calls.append((mycli, args))
        return 13

    monkeypatch.setattr(cli_runner, 'main_batch_without_progress_bar', fake_main_batch_without_progress_bar)

    with pytest.raises(SystemExit) as excinfo:
        cli_runner.run_from_cli_args(cli_args, lambda **_kwargs: client)

    assert excinfo.value.code == 13
    assert batch_calls == [(client, cli_args)]
    assert client.close_called is True


def test_run_from_cli_args_stdin_batch_exits_with_mode_result(monkeypatch: pytest.MonkeyPatch) -> None:
    cli_args = make_cli_args()
    client = DummyMyCli()
    batch_calls: list[tuple[DummyMyCli, main.CliArgs]] = []
    monkeypatch.setattr(main, 'preprocess_cli_args', lambda args, scheme_validator: 0)
    monkeypatch.setattr(cli_runner.sys, 'stdin', SimpleNamespace(isatty=lambda: False))
    monkeypatch.setattr(cli_runner.sys.stderr, 'isatty', lambda: False)

    def fake_main_batch_from_stdin(mycli: DummyMyCli, args: main.CliArgs) -> int:
        batch_calls.append((mycli, args))
        return 14

    monkeypatch.setattr(cli_runner, 'main_batch_from_stdin', fake_main_batch_from_stdin)

    with pytest.raises(SystemExit) as excinfo:
        cli_runner.run_from_cli_args(cli_args, lambda **_kwargs: client)

    assert excinfo.value.code == 14
    assert batch_calls == [(client, cli_args)]
    assert client.close_called is True



from __future__ import annotations

import sys
from typing import Sequence

from mycli.constants import EMPTY_PASSWORD_FLAG_SENTINEL


def filtered_sys_argv() -> Sequence[str | int]:
    args: Sequence[str | int] = sys.argv[1:]
    password_flag_forms = ['-p', '--pass', '--password']

    if args == ['-h']:
        args = ['--help']

    if args and args[-1] in password_flag_forms:
        args = list(args) + [EMPTY_PASSWORD_FLAG_SENTINEL]

    return list(args)


def is_valid_connection_scheme(text: str) -> tuple[bool, str | None]:
    # exit early if the text does not resemble a DSN URI
    if "://" not in text:
        return False, None
    scheme = text.split("://")[0].split("+")[0]
    if scheme in ("mysql", "mysqlx"):
        return True, None
    return False, scheme



from __future__ import annotations

import base64
from datetime import datetime
from decimal import Decimal
from io import TextIOWrapper
import itertools
import os
import shlex
import shutil
from typing import Generator, Literal, Protocol

from cli_helpers.tabular_output import TabularOutputFormatter, preprocessors
from cli_helpers.tabular_output.output_formatter import MISSING_VALUE as DEFAULT_MISSING_VALUE
from cli_helpers.utils import strip_ansi
import click
from configobj import ConfigObj
import prompt_toolkit
from prompt_toolkit.formatted_text import (
    ANSI,
    HTML,
    AnyFormattedText,
    FormattedText,
    to_formatted_text,
    to_plain_text,
)
from prompt_toolkit.shortcuts import PromptSession
from prompt_toolkit.styles.style import _MergedStyle
from pygments.style import Style as PygmentsStyle
from pymysql.cursors import Cursor

from mycli.compat import WIN, is_windows_console
from mycli.constants import DEFAULT_HEIGHT, DEFAULT_WIDTH
import mycli.main_modes.repl as repl_mode
from mycli.packages import special
from mycli.packages.sqlresult import SQLResult
from mycli.packages.tabular_output import sql_format
from mycli.sqlexecute import FIELD_TYPES


class MyCliState(Protocol):
    # Provided by OutputMixin itself; declared so cross-method calls type-check.
    def log_output(self, output: str | AnyFormattedText) -> None: ...
    def get_output_margin(self, status: str | None = None) -> int: ...
    def get_reserved_space(self) -> int: ...


class OutputMixin(MyCliState):
    prompt_lines: int
    multiline_continuation_char: str
    multiplex_pane_title_format: str
    multiplex_window_title_format: str
    terminal_tab_title_format: str
    terminal_window_title_format: str
    toolbar_format: str
    explorer_formatter: TabularOutputFormatter
    explorer_command: str
    explorer_trim_footer: bool
    redirect_formatter: TabularOutputFormatter
    config: ConfigObj
    logfile: TextIOWrapper | Literal[False] | None
    prompt_session: PromptSession | None
    prompt_format: str
    explicit_pager: bool
    ptoolkit_style: _MergedStyle
    helpers_style: PygmentsStyle
    helpers_warnings_style: PygmentsStyle
    main_formatter: TabularOutputFormatter

    def output_timing(self, timing: str, is_warnings_style: bool = False) -> None:
        self.log_output(timing)
        add_style = 'class:warnings.timing' if is_warnings_style else 'class:output.timing'
        formatted_timing = FormattedText([('', timing)])
        styled_timing = to_formatted_text(formatted_timing, style=add_style)
        prompt_toolkit.print_formatted_text(styled_timing, style=self.ptoolkit_style)

    def log_query(self, query: str) -> None:
        if isinstance(self.logfile, TextIOWrapper):
            self.logfile.write(f"\n# {datetime.now()}\n")
            self.logfile.write(query)
            self.logfile.write("\n")

    def log_output(self, output: str | AnyFormattedText) -> None:
        """Log the output in the audit log, if it's enabled."""
        if isinstance(output, (ANSI, HTML, FormattedText)):
            output = to_plain_text(output)
        if isinstance(self.logfile, TextIOWrapper):
            click.echo(output, file=self.logfile)

    def echo(self, s: str, **kwargs) -> None:
        """Print a message to stdout."""
        self.log_output(s)
        click.secho(s, **kwargs)

    def get_output_margin(self, status: str | None = None) -> int:
        """Get the output margin for prompt, footer, timing, and status."""
        if not self.prompt_lines:
            if self.prompt_session and self.prompt_session.app:
                render_counter = self.prompt_session.app.render_counter
            else:
                render_counter = 0
            prompt_string = repl_mode.render_prompt_string(self, self.prompt_format, render_counter)
            self.prompt_lines = to_plain_text(prompt_string).count('\n') + 1
        margin = self.get_reserved_space() + self.prompt_lines
        if special.is_timing_enabled():
            margin += 1
        if status:
            margin += 1 + status.count("\n")

        return margin

    def output(
        self,
        output: itertools.chain[str],
        result: SQLResult,
        is_warnings_style: bool = False,
    ) -> None:
        """Output text to stdout or a pager command."""
        prompt_output = self.prompt_session.output if self.prompt_session is not None else None
        if result.image is not None and not is_windows_console(prompt_output):
            if result.image_protocol == 'iterm2':
                click.secho('')
                self.output_iterm2_image(result.image)
            elif result.image_protocol == 'kitty':
                click.secho('')
                self.output_kitty_image(result.image)
        if output:
            if self.prompt_session is not None:
                size = self.prompt_session.output.get_size()
                size_columns = size.columns
                size_rows = size.rows
            else:
                size_columns = DEFAULT_WIDTH
                size_rows = DEFAULT_HEIGHT

            margin = self.get_output_margin(result.status_plain)

            fits = True
            buf = []
            output_via_pager = self.explicit_pager and special.is_pager_enabled()
            for i, line in enumerate(output, 1):
                self.log_output(line)
                special.write_tee(line)
                special.write_once(line)
                special.write_pipe_once(line)

                if special.is_redirected():
                    pass
                elif special.is_explorer_output():
                    buf.append(line)
                elif fits or output_via_pager:
                    buf.append(line)
                    if len(line) > size_columns or i > (size_rows - margin):
                        fits = False
                        if not self.explicit_pager and special.is_pager_enabled():
                            output_via_pager = True

                        if not output_via_pager:
                            for buf_line in buf:
                                click.secho(buf_line)
                            buf = []
                else:
                    click.secho(line)

            if buf:

                def newlinewrapper(text: list[str]) -> Generator[str, None, None]:
                    for line in text:
                        yield line + "\n"

                if special.is_explorer_output():
                    if self.explorer_exists() or not self.explorer_command:
                        old_pager = os.environ.get('PAGER')
                        os.environ['PAGER'] = self.explorer_command or old_pager or 'less'
                        click.echo_via_pager(newlinewrapper(buf))
                        if old_pager:
                            os.environ['PAGER'] = old_pager
                        else:
                            del os.environ['PAGER']
                    else:
                        click.secho(f'Configured explorer command not found: {self.explorer_command}.', err=True, fg='red')
                elif output_via_pager:
                    click.echo_via_pager(newlinewrapper(buf))
                else:
                    for line in buf:
                        click.secho(line)

        if result.status:
            self.log_output(result.status_plain)
            add_style = 'class:warnings.status' if is_warnings_style else 'class:output.status'
            if isinstance(result.status, FormattedText):
                status = result.status
            else:
                status = FormattedText([('', result.status_plain)])
            styled_status = to_formatted_text(status, style=add_style)
            prompt_toolkit.print_formatted_text(styled_status, style=self.ptoolkit_style)

    def output_iterm2_image(self, image: bytes) -> None:
        """Emit a PNG using the iTerm2 inline image protocol."""
        filename = base64.b64encode(b'chart.png').decode('ascii')
        contents = base64.b64encode(image).decode('ascii')
        click.echo(f'\x1b]1337;File=name={filename};size={len(image)};width=auto;height=auto;preserveAspectRatio=1;inline=1:{contents}\x07')

    def output_kitty_image(self, image: bytes) -> None:
        """Emit a PNG using Kitty's direct graphics protocol."""
        contents = base64.b64encode(image).decode('ascii')
        chunks = [contents[index : index + 4096] for index in range(0, len(contents), 4096)] or ['']
        for index, chunk in enumerate(chunks):
            action = 'a=T,f=100,' if index == 0 else ''
            more = 1 if index < len(chunks) - 1 else 0
            click.echo(f'\x1b_G{action}m={more};{chunk}\x1b\\', nl=False)
        click.echo()

    def configure_pager(self) -> None:
        if not os.environ.get("LESS"):
            os.environ["LESS"] = "-RXF"

        config_pager = self.config["main"]["pager"]

        if WIN and config_pager == 'less' and not shutil.which(config_pager):
            config_pager = 'more'

        if config_pager:
            special.set_pager(config_pager)
            self.explicit_pager = True
        else:
            self.explicit_pager = False

        if not self.config["main"].as_bool("enable_pager"):
            special.disable_pager()

    def explorer_exists(self) -> bool:
        if cmd := shlex.split(self.explorer_command or ''):
            return bool(shutil.which(cmd[0]))
        return False

    def format_sqlresult(
        self,
        result,
        is_expanded: bool = False,
        is_redirected: bool = False,
        null_string: str | None = None,
        numeric_alignment: str = 'right',
        binary_display: str | None = None,
        max_width: int | None = None,
        is_warnings_style: bool = False,
    ) -> itertools.chain[str]:
        if is_redirected:
            use_formatter = self.redirect_formatter
        elif special.is_explorer_output():
            use_formatter = self.explorer_formatter
        else:
            use_formatter = self.main_formatter

        is_expanded = is_expanded or use_formatter.format_name == "vertical"
        output: itertools.chain[str] = itertools.chain()

        output_kwargs = {
            "dialect": "unix",
            "disable_numparse": True,
            "preserve_whitespace": True,
            "style": self.helpers_warnings_style if is_warnings_style else self.helpers_style,
        }
        default_kwargs = use_formatter._output_formats[use_formatter.format_name].formatter_args

        if null_string is not None and default_kwargs.get('missing_value') == DEFAULT_MISSING_VALUE:
            output_kwargs['missing_value'] = null_string

        if use_formatter.format_name not in sql_format.supported_formats and binary_display != 'utf8':
            output_kwargs["preprocessors"] = (preprocessors.convert_to_undecoded_string,)

        if result.preamble:
            output = itertools.chain(output, [result.preamble])

        if result.header or (result.rows and result.preamble):
            column_types = None
            colalign = None
            if isinstance(result.rows, Cursor):

                def get_col_type(col) -> type:
                    col_type = FIELD_TYPES.get(col[1], str)
                    return col_type if type(col_type) is type else str

                if result.rows.rowcount > 0:
                    column_types = [get_col_type(tup) for tup in result.rows.description]
                    colalign = [numeric_alignment if x in (int, float, Decimal) else 'left' for x in column_types]
                else:
                    column_types, colalign = [], []

            if max_width is not None and isinstance(result.rows, Cursor):
                result_rows = list(result.rows)
            else:
                result_rows = result.rows

            formatted = use_formatter.format_output(
                result_rows,
                result.header or [],
                format_name="vertical" if is_expanded else None,
                column_types=column_types,
                colalign=colalign,
                **output_kwargs,
            )

            if isinstance(formatted, str):
                formatted = formatted.splitlines()

            if special.is_explorer_output() and self.explorer_trim_footer:
                formatted = list(formatted)
                formatted.pop()

            formatted = iter(formatted)

            if not is_expanded and max_width and result.header and result_rows:
                first_line = next(formatted)
                if len(strip_ansi(first_line)) > max_width:
                    formatted = use_formatter.format_output(
                        result_rows,
                        result.header,
                        format_name="vertical",
                        column_types=column_types,
                        **output_kwargs,
                    )
                    if isinstance(formatted, str):
                        formatted = iter(formatted.splitlines())
                else:
                    formatted = itertools.chain([first_line], formatted)

            output = itertools.chain(output, formatted)

        if result.postamble:
            output = itertools.chain(output, [result.postamble])

        return output

    def get_reserved_space(self) -> int:
        """Get the number of lines to reserve for the completion menu."""
        reserved_space_ratio = 0.45
        max_reserved_space = 8
        _, height = shutil.get_terminal_size()
        return min(int(round(height * reserved_space_ratio)), max_reserved_space)



from collections import namedtuple
from typing import Literal

# Query tuples are used for maintaining history
Query = namedtuple("Query", ["query", "successful", "mutating"])

OutputMode = Literal[
    'explorer',
    'expanded',
    'tabular',
]

ImageProtocol = Literal[
    'none',
    'iterm2',
    'kitty',
]
